from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from sklearn.utils import check_random_state

from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
)
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES, SUPPORTED_TARGET_TYPES
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.pipeline.util import get_dataset

from pytest import fixture

from test.conftest import DEFAULT_SEED

Data = Tuple[SUPPORTED_FEAT_TYPES, SUPPORTED_TARGET_TYPES]


def astype(
    t: np.ndarray | list | csr_matrix | pd.DataFrame | pd.Series,
    x: Any,
) -> Any:
    """Convert data to allowed types"""
    if t == np.ndarray:
        return np.asarray(x)
    else:
        return t(x)  # type: ignore


# TODO Remove the implementation in autosklearn.pipeline.util and just put here
@fixture
def make_sklearn_dataset() -> Callable:
    """
    Parameters
    ----------
    name : str = "iris"
        Name of the dataset to get

    make_sparse : bool = False
        Wehther to make the data sparse

    add_NaNs : bool = False
        Whether to add NaNs to the data

    train_size_maximum : int = 150
        THe maximum size of training data

    make_multilabel : bool = False
        Whether to force the data into being multilabel

    make_binary : bool = False
        Whether to force the data into being binary

    task: Optional[int] = None
        The task of the data, required for the datamanager

    feat_type: Optional[Dict | str] = None
        The features types for the data if making a XYDataManager

    as_datamanager: bool = False
        Wether to return the information as an XYDataManager

    Returns
    -------
    (X_train, Y_train, X_test, Y_Test) | XYDataManager
    """

    def _make(
        name: str = "iris",
        make_sparse: bool = False,
        add_NaNs: bool = False,
        train_size_maximum: int = 150,
        make_multilabel: bool = False,
        make_binary: bool = False,
        task: Optional[int] = None,
        feat_type: Optional[Dict | str] = None,
        as_datamanager: bool = False,
        return_target_as_string: bool = False,
    ) -> Any:
        X, y, Xt, yt = get_dataset(
            dataset=name,
            make_sparse=make_sparse,
            add_NaNs=add_NaNs,
            train_size_maximum=train_size_maximum,
            make_multilabel=make_multilabel,
            make_binary=make_binary,
            return_target_as_string=return_target_as_string,
        )

        if not as_datamanager:
            return (X, y, Xt, yt)
        else:

            assert task is not None and feat_type is not None
            if isinstance(feat_type, str):
                feat_type = {i: feat_type for i in range(X.shape[1])}

            return XYDataManager(
                X,
                y,
                Xt,
                yt,
                task=task,
                dataset_name=name,
                feat_type=feat_type,
            )

    return _make


def _make_binary_data(
    dims: Tuple[int, ...] = (100, 3),
    weights: Optional[Sequence[float] | np.ndarray] = None,
    types: Tuple[
        np.ndarray | csr_matrix | pd.DataFrame | list,
        np.ndarray | csr_matrix | pd.DataFrame | list | pd.Series,
    ] = (np.ndarray, np.ndarray),
    random_state: int | np.random.RandomState = DEFAULT_SEED,
) -> Data:
    X_type, y_type = types
    rs = check_random_state(random_state)

    classes = [0, 1]

    if not weights:
        weights = np.ones_like(classes) / len(classes)

    assert len(weights) == len(classes)
    weights = weights / np.sum(weights, keepdims=True)

    X = rs.rand(*dims)
    y = rs.choice([0, 1], dims[0], p=weights)

    return astype(X_type, X), astype(y_type, y)


def _make_multiclass_data(
    dims: Tuple[int, ...] = (100, 3),
    classes: int | np.ndarray | List = 3,
    weights: Optional[np.ndarray | List[float]] = None,
    types: Tuple[
        np.ndarray | csr_matrix | pd.DataFrame | list,
        np.ndarray | csr_matrix | pd.DataFrame | list | pd.Series,
    ] = (np.ndarray, np.ndarray),
    random_state: int | np.random.RandomState = DEFAULT_SEED,
) -> Data:
    X_type, y_type = types

    if isinstance(classes, int):
        classes = np.asarray(list(range(classes)))

    rs = check_random_state(random_state)

    if not weights:
        weights = np.ones_like(classes) / len(classes)

    assert len(weights) == len(classes)
    weights = weights / np.sum(weights, keepdims=True)

    X = rs.rand(*dims)
    y = rs.choice(classes, dims[0], p=weights)

    return astype(X_type, X), astype(y_type, y)


def _make_multilabel_data(
    dims: Tuple[int, ...] = (100, 3),
    classes: np.ndarray | List = [[0, 0], [0, 1], [1, 0], [1, 1]],
    weights: Optional[np.ndarray | List[float]] = None,
    types: Tuple[
        np.ndarray | csr_matrix | pd.DataFrame | list,
        np.ndarray | csr_matrix | pd.DataFrame | list | pd.Series,
    ] = (np.ndarray, np.ndarray),
    random_state: int | np.random.RandomState = DEFAULT_SEED,
) -> Data:
    X_type, y_type = types

    classes = np.asarray(classes)
    assert classes.ndim > 1 and classes.shape[1] > 1

    rs = check_random_state(random_state)

    # Weights indicate each label tuple, and not the weights of individual labels
    # in that tuple
    if not weights:
        weights = np.ones(classes.shape[0]) / len(classes)

    assert len(weights) == len(classes)
    weights = weights / np.sum(weights, keepdims=True)

    X = rs.rand(*dims)

    class_indices = rs.choice(len(classes), dims[0], p=weights)
    y = classes[class_indices]

    return astype(X_type, X), astype(y_type, y)


def _make_regression_data(
    dims: Tuple[int, ...] = (100, 3),
    types: Tuple[
        np.ndarray | csr_matrix | pd.DataFrame | list,
        np.ndarray | csr_matrix | pd.DataFrame | list | pd.Series,
    ] = (np.ndarray, np.ndarray),
    random_state: int | np.random.RandomState = DEFAULT_SEED,
) -> Data:
    X_type, y_type = types
    rs = check_random_state(random_state)

    if X_type == csr_matrix:
        X = rs.choice([0, 1], dims)
    else:
        X = rs.rand(*dims)

    y = rs.rand(dims[0])

    return astype(X_type, X), astype(y_type, y)


def _make_multioutput_regression_data(
    dims: Tuple[int, ...] = (100, 3),
    targets: int = 2,
    types: Tuple[
        np.ndarray | csr_matrix | pd.DataFrame | list,
        np.ndarray | csr_matrix | pd.DataFrame | list | pd.Series,
    ] = (np.ndarray, np.ndarray),
    random_state: int | np.random.RandomState = DEFAULT_SEED,
) -> Data:
    X_type, y_type = types

    rs = check_random_state(random_state)

    if X_type == csr_matrix:
        X = rs.choice([0, 1], dims)
    else:
        X = rs.rand(*dims)

    y = rs.rand(dims[0], targets)

    return astype(X_type, X), astype(y_type, y)


@fixture
def make_data() -> Callable[..., Data]:
    """Generate some arbitrary x,y data

    Parameters
    ----------
    kind: int = BINARY_CLASSIFICATION
        The task type, one of BINARY_CLASSIFICATION, MULTICLASS_CLASSIFICATION, ...

    **kwargs: Any
        See the corresponding `_make_<x>`

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The generated data
    """

    def _make(
        kind: int = BINARY_CLASSIFICATION,
        **kwargs: Any,
    ) -> Data:
        dispatches = {
            BINARY_CLASSIFICATION: _make_binary_data,
            MULTICLASS_CLASSIFICATION: _make_multiclass_data,
            MULTILABEL_CLASSIFICATION: _make_multilabel_data,
            REGRESSION: _make_regression_data,
            MULTIOUTPUT_REGRESSION: _make_multioutput_regression_data,
        }

        f = dispatches[kind]
        return f(**kwargs)  # type: ignore

    return _make



import abc
from typing import Any, Dict, Union

import numpy as np
import scipy.sparse

from autosklearn.pipeline.components.data_preprocessing.feature_type import (
    FeatTypeSplit,
)


class AbstractDataManager:
    __metaclass__ = abc.ABCMeta

    def __init__(self, name: str):

        self._data = dict()  # type: Dict
        self._info = dict()  # type: Dict
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def data(self) -> Dict[str, np.ndarray]:
        return self._data

    @property
    def info(self) -> Dict[str, Any]:
        return self._info

    @property
    def feat_type(self) -> Dict[Union[str, int], str]:
        return self._feat_type

    @feat_type.setter
    def feat_type(self, value: Dict[Union[str, int], str]) -> None:
        self._feat_type = value

    @property
    def encoder(self) -> FeatTypeSplit:
        return self._encoder

    @encoder.setter
    def encoder(self, value: FeatTypeSplit) -> FeatTypeSplit:
        self._encoder = value

    def __repr__(self) -> str:
        return "DataManager : " + self.name

    def __str__(self) -> str:
        val = "DataManager : " + self.name + "\ninfo:\n"
        for item in self.info:
            val = val + "\t" + item + " = " + str(self.info[item]) + "\n"
        val = val + "data:\n"

        for subset in self.data:
            val = val + "\t%s = %s %s %s\n" % (
                subset,
                type(self.data[subset]),
                str(self.data[subset].shape),
                str(self.data[subset].dtype),
            )
            if isinstance(self.data[subset], scipy.sparse.spmatrix):
                val = val + "\tdensity: %f\n" % (
                    float(len(self.data[subset].data))
                    / self.data[subset].shape[0]
                    / self.data[subset].shape[1]
                )
        val = val + "feat_type:\t" + str(self.feat_type) + "\n"
        return val



import numpy as np
import sklearn.metrics

from autosklearn.constants import REGRESSION
from autosklearn.ensembles.multiobjective_dummy_ensemble import (
    MultiObjectiveDummyEnsemble,
)
from autosklearn.ensembles.singlebest_ensemble import SingleModelEnsemble
from autosklearn.metrics import MAXINT, make_scorer, root_mean_squared_error


def test_MultiObjectiveDummyEnsemble(backend):
    negative_root_mean_squared_error = make_scorer(
        "negative_root_mean_squared_error",
        sklearn.metrics.mean_squared_error,
        optimum=0,
        worst_possible_result=-MAXINT,
        greater_is_better=True,
        squared=False,
    )

    ensemble = MultiObjectiveDummyEnsemble(
        task_type=REGRESSION,
        metrics=[root_mean_squared_error, negative_root_mean_squared_error],
        random_state=0,
        backend=backend,
    )

    # fit the whole thing
    X_data = np.random.random(size=(100, 2))
    y_true = np.full((100), 5.5)
    predictions = []
    for i in range(20, 1, -1):
        pred = np.ones((100), dtype=np.float32) * i
        predictions.append(pred)
    identifiers = [(i, i, 0.0) for i in range(20)]

    ensemble.fit(
        base_models_predictions=predictions,
        X_data=X_data,
        true_targets=y_true,
        model_identifiers=identifiers,
        runs=[],
    )

    # Because the target value is 5.5 some of the predictions have the same error
    # -> not every prediction is on the Pareto front
    assert len(ensemble.pareto_set_) == 15
    for sub_ensemble in ensemble.pareto_set_:
        assert isinstance(sub_ensemble, SingleModelEnsemble)

    # test the prediction
    ensemble_prediction = np.random.random((1, 100, 2))
    prediction = ensemble.predict(ensemble_prediction)
    assert prediction.shape == (100, 2)

    assert str(ensemble) == """MultiObjectiveDummyEnsemble: 15 models"""

    # From here on everything is supposed to return the output of the best
    # ensemble according to the 1st metric
    models_with_weights = ensemble.get_models_with_weights(
        {identifier: i for i, identifier in enumerate(identifiers)}
    )
    assert models_with_weights == [(1.0, 14)]

    identifiers_with_weights = ensemble.get_identifiers_with_weights()
    assert identifiers_with_weights == [((14, 14, 0.0), 1.0)]

    selected_model_identifiers = ensemble.get_selected_model_identifiers()
    assert selected_model_identifiers == [(14, 14, 0.0)]

    best_model_score = ensemble.get_validation_performance()
    assert best_model_score == 0.5



from __future__ import annotations

from typing import Iterator, Mapping, Optional, Tuple

import sys
import time
from contextlib import contextmanager
from itertools import repeat

from typing_extensions import Literal


class TimingTask:
    """A task to start"""

    def __init__(self, name: str):
        """
        Parameters
        ----------
        name : str
            The name of the task
        """
        self.name = name
        self.cpu_start: Optional[float] = None
        self.cpu_end: Optional[float] = None
        self.wall_start: Optional[float] = None
        self.wall_end: Optional[float] = None

    def start(self) -> None:
        """Start the task"""
        if not self.started():
            self.cpu_start = time.process_time()
            self.wall_start = time.time()

    def stop(self) -> None:
        """Stop the task"""
        if not self.finished():
            self.cpu_end = time.process_time()
            self.wall_end = time.time()

    def finished(self) -> bool:
        """Whether this task has been finished"""
        return self.cpu_end is not None

    def started(self) -> bool:
        """Whether this task has been started"""
        return self.cpu_start is not None

    @property
    def cpu_duration(self) -> Optional[float]:
        """The duration according to cpu clock time"""
        if self.cpu_start and self.cpu_end:
            return self.cpu_end - self.cpu_start
        else:
            return None

    @property
    def wall_duration(self) -> Optional[float]:
        """The duration according to wall clock time"""
        if self.wall_end and self.wall_start:
            return self.wall_end - self.wall_start
        else:
            return None

    @property
    def durations(self) -> Tuple[Optional[float], Optional[float]]:
        """The durations (cpu, wall)"""
        return self.cpu_duration, self.wall_duration


class StopWatch(Mapping[str, TimingTask]):
    """Class to collect timing tasks."""

    def __init__(self) -> None:
        self.tasks = {"_stopwatch_": TimingTask("_stopwatch_")}
        self.tasks["_stopwatch_"].start()

    def __contains__(self, name: object) -> bool:
        return name in self.tasks

    def __getitem__(self, name: str) -> TimingTask:
        return self.tasks[name]

    def __len__(self) -> int:
        return len(self.tasks)

    def __iter__(self) -> Iterator[str]:
        return iter(self.tasks)

    def start(self, name: str) -> None:
        """Start a given task with a name"""
        task = TimingTask(name)
        self.tasks[name] = task
        task.start()

    def stop(self, name: str) -> None:
        """Stop a given task"""
        if name in self.tasks:
            self.tasks[name].stop()
        else:
            sys.stderr.write(f"No task with name {name}")

    def total_cpu(self) -> float:
        """Return sum of CPU time for all so far finished tasks."""
        tasks = self.tasks.values()
        return sum([t.cpu_duration for t in tasks if t.cpu_duration is not None])

    def total_wall(self) -> float:
        """Return sum of wall clock time for all so far finished tasks."""
        tasks = self.tasks.values()
        return sum([t.wall_duration for t in tasks if t.wall_duration is not None])

    def time_since(
        self,
        name: str,
        phase: Literal["start", "end"] = "start",
        default: float = 0.0,
        raises: bool = False,
    ) -> float:
        """The wall clock time since a task either began or ended

        Parameters
        ----------
        name : str
            The name of the task

        phase : Literal["start", "end"] = "start"
            From which phase you want to know the time elapsed since

        default: float = 0.0
            If None (default) then an error is raised if an answer can't be given.

        raises: bool = False
            Whether the method should raise if it can't find a valid time

        Returns
        -------
        float
            The time elapsed

        Raises
        ------
        ValueError
            If no default is specified and
            * the task has not been registered
            * the "start" and the task never started
            * the "end" and the task never started
        """
        task = self.tasks.get(name, None)

        if task is None:
            if raises:
                raise ValueError(f"Task not listed in {list(self.tasks.keys())}")
            else:
                return default

        if phase == "start":
            event_time = task.wall_start
        elif phase == "end":
            event_time = task.wall_end
        else:
            raise NotImplementedError()

        if event_time is None:
            if raises:
                raise ValueError(f"Task {task} has no time for {phase}")
            else:
                return default

        return time.time() - event_time

    def wall_elapsed(self, name: str) -> float:
        """Get the currently elapsed wall time for a task"""
        if name not in self.tasks:
            return 0.0

        task = self.tasks[name]
        if task.wall_start is None:
            return 0.0

        if task.wall_duration is not None:
            return task.wall_duration
        else:
            return time.time() - task.wall_start

    def cpu_elapsed(self, name: str) -> float:
        """Get the currently elapsed cpu time for a task"""
        if name not in self.tasks:
            return 0.0

        task = self.tasks[name]
        if task.cpu_start is None:
            return 0.0

        if task.cpu_duration is not None:
            return task.cpu_duration
        else:
            return time.time() - task.cpu_start

    @contextmanager
    def time(self, name: str) -> Iterator[TimingTask]:
        """Start timing a task

        Parameters
        ----------
        name : str
            The name of the task to measure
        """
        task = TimingTask(name)
        self.tasks[name] = task
        task.start()
        yield task
        task.stop()

    def __str__(self) -> str:
        headers = [
            "Name",
            "CPUStart",
            "CPUEnd",
            "CPUDur",
            "WallStart",
            "WallEnd",
            "WallDur",
        ]
        header = "|".join([f"{h:10s}" for h in headers])

        sep = "-" * 12
        seperator = "+" + "+".join(repeat(sep, len(headers))) + "+"

        entries = []
        for name, task in self.tasks.items():
            entry = (
                f"{task.name:10s} | {task.cpu_start:10.5f} | {task.cpu_end: 10.5f} | "
                f"{task.cpu_duration:10.5f} | {task.wall_start: 10s} | "
                f"{task.wall_end:10s} | {task.wall_duration:10s} |"
            )
            entries.append(entry)

        return "\n".join([header, seperator] + entries)



from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import warnings

import numpy as np
import pandas as pd
from scipy.sparse import spmatrix
from sklearn.model_selection import train_test_split

from autosklearn.evaluation.splitter import CustomStratifiedShuffleSplit

# TODO: TypedDict with python 3.8
#
#   When upgrading to python 3.8 as minimum version, this should be a TypedDict
#   so that mypy can identify the fields types
DatasetCompressionSpec = Dict[str, Union[float, List[str]]]

# Default specification for arg `dataset_compression`
default_dataset_compression_arg: DatasetCompressionSpec = {
    "memory_allocation": 0.1,
    "methods": ["precision", "subsample"],
}


def validate_dataset_compression_arg(
    dataset_compression: Mapping[str, Any], memory_limit: int
) -> DatasetCompressionSpec:
    """Validates and return a correct dataset_compression argument

    The returned value can be safely used with `reduce_dataset_size_if_too_large`.

    Parameters
    ----------
    dataset_compression: Mapping[str, Any]
        The arg to validate

    Returns
    -------
    DatasetCompressionSpec
        The validated and correct dataset compression spec
    """
    if isinstance(dataset_compression, Mapping):
        # Fill with defaults if they don't exist
        dataset_compression = {**default_dataset_compression_arg, **dataset_compression}

        parsed_keys = set(dataset_compression.keys())
        default_keys = set(default_dataset_compression_arg.keys())

        # Must contain known keys
        if parsed_keys != default_keys:
            raise ValueError(
                f"Unknown key(s) in ``dataset_compression``, {parsed_keys}."
                f"\nPossible keys are {default_keys}"
            )

        memory_allocation = dataset_compression["memory_allocation"]

        # "memory_allocation" must be float or int
        if not (
            isinstance(memory_allocation, float) or isinstance(memory_allocation, int)
        ):
            raise ValueError(
                "key 'memory_allocation' must be an `int` or `float`"
                f"\ntype = {memory_allocation}"
                f"\ndataset_compression = {dataset_compression}"
            )

        # "memory_allocation" must be in (0,1) if float
        if isinstance(memory_allocation, float) and not (0.0 < memory_allocation < 1.0):
            raise ValueError(
                "key 'memory_allocation' if float must be in (0, 1)"
                f"\nmemory_allocation = {memory_allocation}"
                f"\ndataset_compression = {dataset_compression}"
            )

        # "memory_allocation" if absolute, should be > 0 and < memory_limit
        if isinstance(memory_allocation, int) and not (
            0 < memory_allocation < memory_limit
        ):
            raise ValueError(
                f"key 'memory_allocation' if int must be in (0, {memory_limit})"
                f"\nmemory_allocation = {memory_allocation}"
                f"\ndataset_compression = {dataset_compression}"
            )

        # "methods" must be non-empty sequence
        if (
            not isinstance(dataset_compression["methods"], Sequence)
            or len(dataset_compression["methods"]) <= 0
        ):
            raise ValueError(
                "key 'methods' must be a non-empty list"
                f"\nmethods = {dataset_compression['methods']}"
                f"\ndataset_compression = {dataset_compression}"
            )

        # "methods" must contain known methods
        if any(
            method
            not in cast(Sequence, default_dataset_compression_arg["methods"])  # mypy
            for method in dataset_compression["methods"]
        ):
            valid_methods = default_dataset_compression_arg["methods"]
            raise ValueError(
                f"key 'methods' can only contain {valid_methods}"
                f"\nmethods = {dataset_compression['methods']}"
                f"\ndataset_compression = {dataset_compression}"
            )

        return cast(DatasetCompressionSpec, dataset_compression)
    else:
        raise ValueError(
            f"Unknown type for `dataset_compression` {type(dataset_compression)}"
            f"\ndataset_compression = {dataset_compression}"
        )


class _DtypeReductionMapping(Mapping):
    """
    Unfortuantly, mappings compare by hash(item) and not the __eq__ operator
    between the key and the item.

    Hence we wrap the dict in a Mapping class and implement our own __getitem__
    such that we do use __eq__ between keys and query items.

    >>> np.float32 == dtype('float32') # True, they are considered equal
    >>>
    >>> mydict = { np.float32: 'hello' }
    >>>
    >>> # Equal by __eq__ but dict operations fail
    >>> np.dtype('float32') in mydict # False
    >>> mydict[dtype('float32')]  # KeyError

    This mapping class fixes that supporting the `in` operator as well as `__getitem__`

    >>> reduction_mapping = _DtypeReductionMapping()
    >>>
    >>> reduction_mapping[np.dtype('float64')] # np.float32
    >>> np.dtype('float32') in reduction_mapping # True
    """

    # Information about dtype support
    _mapping: Dict[type, type] = {
        np.float32: np.float32,
        np.float64: np.float32,
    }

    # In spite of the names, np.float96 and np.float128
    # provide only as much precision as np.longdouble,
    # that is, 80 bits on most x86 machines and 64 bits
    # in standard Windows builds.
    if hasattr(np, "float96"):
        _mapping[np.float96] = np.float64

    if hasattr(np, "float128"):
        _mapping[np.float128] = np.float64

    @classmethod
    def __getitem__(cls, item: type) -> type:
        for k, v in cls._mapping.items():
            if k == item:
                return v
        raise KeyError(item)

    @classmethod
    def __iter__(cls) -> Iterator[type]:
        return iter(cls._mapping.keys())

    @classmethod
    def __len__(cls) -> int:
        return len(cls._mapping)


reduction_mapping = _DtypeReductionMapping()
supported_precision_reductions = list(reduction_mapping)


def binarization(array: Union[List, np.ndarray]) -> np.ndarray:
    # Takes a binary-class datafile and turn the max value (positive class)
    # into 1 and the min into 0
    array = np.array(array, dtype=float)  # conversion needed to use np.inf
    if len(np.unique(array)) > 2:
        raise ValueError(
            "The argument must be a binary-class datafile. "
            "{} classes detected".format(len(np.unique(array)))
        )

    # manipulation which aims at avoid error in data
    # with for example classes '1' and '2'.
    array[array == np.amax(array)] = np.inf
    array[array == np.amin(array)] = 0
    array[array == np.inf] = 1
    return np.array(array, dtype=int)


def multilabel_to_multiclass(array: Union[List, np.ndarray]) -> np.ndarray:
    array = binarization(array)
    return np.array([np.nonzero(array[i, :])[0][0] for i in range(len(array))])


def convert_to_num(Ybin: np.ndarray) -> np.ndarray:
    """
    Convert binary targets to numeric vector
    typically classification target values
    :param Ybin:
    :return:
    """
    result = np.array(Ybin)
    if len(Ybin.shape) != 1:
        result = np.dot(Ybin, range(Ybin.shape[1]))
    return result


def convert_to_bin(Ycont: List, nval: int, verbose: bool = True) -> List:
    # Convert numeric vector to binary (typically classification target values)
    if verbose:
        pass
    Ybin = [[0] * nval for _ in range(len(Ycont))]
    for i in range(len(Ybin)):
        line = Ybin[i]
        line[int(Ycont[i])] = 1
        Ybin[i] = line
    return Ybin


def predict_RAM_usage(X: np.ndarray, categorical: List[bool]) -> float:
    # Return estimated RAM usage of dataset after OneHotEncoding in bytes.
    estimated_columns = 0
    for i, cat in enumerate(categorical):
        if cat:
            unique_values = np.unique(X[:, i])
            num_unique_values = np.sum(np.isfinite(unique_values))
            estimated_columns += num_unique_values
        else:
            estimated_columns += 1
    estimated_ram = estimated_columns * X.shape[0] * X.dtype.itemsize
    return estimated_ram


def subsample(
    X: Union[np.ndarray, spmatrix],
    y: np.ndarray,
    is_classification: bool,
    sample_size: Union[float, int],
    random_state: Optional[Union[int, np.random.RandomState]] = None,
) -> Tuple[Union[np.ndarray, spmatrix], np.ndarray]:
    """Subsamples data returning the same type as it recieved.

    If `is_classification`, we split using a stratified shuffle split which
    preserves unique labels in the training set.

    NOTE:
    It's highly unadvisable to use lists here. In order to preserver types,
    we convert to a numpy array and then back to a list.

    NOTE2:
    Interestingly enough, StratifiedShuffleSplut and descendants don't support
    sparse `y` in `split(): _check_array` call. Hence, neither do we.

    Parameters
    ----------
    X: Union[np.ndarray, spmatrix]
        The X's to subsample

    y: np.ndarray
        The Y's to subsample

    is_classification: bool
        Whether this is classification data or regression data. Required for
        knowing how to split.

    sample_size: float | int
        If float, percentage of data to take otherwise if int, an absolute
        count of samples to take.

    random_state: int | RandomState = None
        The random state to pass to the splitted

    Returns
    -------
    (np.ndarray | spmatrix, np.ndarray)
        The X and y subsampled according to sample_size
    """
    if isinstance(X, List):
        X = np.asarray(X)

    if isinstance(y, List):
        y = np.asarray(y)

    if is_classification:
        splitter = CustomStratifiedShuffleSplit(
            train_size=sample_size, random_state=random_state
        )
        left_idxs, _ = next(splitter.split(X=X, y=y))

        # This function supports pandas objects but they won't get here
        # yet as we do not reduce the size of pandas dataframes.
        if isinstance(X, pd.DataFrame):
            idxs = X.index[left_idxs]
            X = X.loc[idxs]
        else:
            X = X[left_idxs]

        if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            idxs = y.index[left_idxs]
            y = y.loc[idxs]
        else:
            y = y[left_idxs]

    else:
        X, _, y, _ = train_test_split(  # type: ignore
            X,
            y,
            train_size=sample_size,
            random_state=random_state,
        )

    return X, y


def reduce_precision(
    X: Union[np.ndarray, spmatrix]
) -> Tuple[Union[np.ndarray, spmatrix], Type]:
    """Reduces the precision of a np.ndarray or spmatrix containing floats

    Parameters
    ----------
    X:  np.ndarray | spmatrix
        The data to reduce precision of.

    Returns
    -------
    (ndarray | spmatrix, dtype)
        Returns the reduced data X along with the dtype it was reduced to.
    """
    if X.dtype not in supported_precision_reductions:
        raise ValueError(
            f"X.dtype = {X.dtype} not equal to any supported"
            f" {supported_precision_reductions}"
        )

    precision = reduction_mapping[X.dtype]
    return X.astype(precision), precision


def reduce_dataset_size_if_too_large(
    X: Union[np.ndarray, spmatrix],
    y: np.ndarray,
    memory_limit: int,
    is_classification: bool,
    random_state: Union[int, np.random.RandomState] = None,
    operations: List[str] = ["precision", "subsample"],
    memory_allocation: Union[int, float] = 0.1,
) -> Tuple[Union[np.ndarray, spmatrix], np.ndarray]:
    f"""Reduces the size of the dataset if it's too close to the memory limit.

    Follows the order of the operations passed in and retains the type of its
    input.

    Precision reduction will only work on the following float types:
    -   {supported_precision_reductions}

    Subsampling will ensure that the memory limit is satisfied while precision reduction
    will only perform one level of precision reduction. Technically, you could supply
    multiple rounds of precision reduction, i.e. to reduce np.float128 to np.float32
    you could use `operations = ['precision'] * 2`.

    However, if that's the use case, it'd be advised to simply use the function
    `autosklearn.util.data.reduce_precision`.

    NOTE: limitations
    * Does not support dataframes yet
        -   Requires implementing column wise precision reduction
        -   Requires calculating memory usage

    Parameters
    ----------
    X: np.ndarray | spmatrix
        The features of the dataset.

    y: np.ndarray
        The labels of the dataset.

    memory_limit: int
        The amount of memory allocated in megabytes

    is_classification: bool
        Whether it's a classificaiton dataset or not. This is important when
        considering how to subsample.

    random_state: int | RandomState = None
        The random_state to use for subsampling.

    operations: List[str] = ['precision', 'subsampling']
        A list of operations that are permitted to be performed to reduce
        the size of the dataset.

        **precision**

        Reduce the precision of float types

        **subsample**

        Reduce the amount of samples of the dataset such that it fits into the allocated
        memory. Ensures stratification and that unique labels are present

    memory_allocation: Union[int, float] = 0.1
        The amount of memory to allocate to the dataset. A float specifys that the
        dataset will try to be fit into that percentage of memory. An int specifies an
        absolute amount.

    Returns
    -------
    (spmatrix | np.ndarray, np.ndarray)
        The reduced X, y if reductions were needed
    """
    # Validation
    assert memory_limit > 0

    if isinstance(memory_allocation, float):
        if not (0.0 < memory_allocation < 1.0):
            raise ValueError("memory_allocation if float must be in (0, 1)")

        allocated_memory = memory_limit * memory_allocation

    elif isinstance(memory_allocation, int):
        if not (0 < memory_allocation < memory_limit):
            raise ValueError("memory_allocation if int must be in (0, memory_limit)")

        allocated_memory = memory_allocation

    else:
        raise ValueError(
            f"Unknown type for `memory_allocation` {type(memory_allocation)}"
        )

    if "precision" in operations and X.dtype not in supported_precision_reductions:
        raise ValueError(f"Unsupported type `{X.dtype}` for precision reduction")

    def megabytes(arr: Union[np.ndarray, spmatrix]) -> float:
        return (arr.nbytes if isinstance(X, np.ndarray) else arr.data.nbytes) / (
            2**20
        )

    for operation in operations:

        if operation == "precision":
            # If the dataset is too big for the allocated memory,
            # we then try to reduce the precision if it's a high precision dataset
            if megabytes(X) > allocated_memory:
                X, precision = reduce_precision(X)
                warnings.warn(
                    f"Dataset too large for allocated memory {allocated_memory}MB, "
                    f"reduced the precision from {X.dtype} to {precision}",
                )

        elif operation == "subsample":
            # If the dataset is still too big such that we couldn't fit
            # into the allocated memory, we subsample it so that it does
            if megabytes(X) > allocated_memory:

                n_samples_before = X.shape[0]
                sample_percentage = allocated_memory / megabytes(X)

                # NOTE: type ignore
                #
                # Tried the generic `def subsample(X: T) -> T` approach but it was
                # failing elsewhere, keeping it simple for now
                X, y = subsample(  # type: ignore
                    X,
                    y,
                    sample_size=sample_percentage,
                    is_classification=is_classification,
                    random_state=random_state,
                )

                n_samples_after = X.shape[0]
                warnings.warn(
                    f"Dataset too large for allocated memory {allocated_memory}MB,"
                    f" reduced number of samples from {n_samples_before} to"
                    f" {n_samples_after}."
                )

        else:
            raise ValueError(f"Unknown operation `{operation}`")

    return X, y



import numpy as np
import scipy.sparse

from autosklearn.pipeline.implementations.MinorityCoalescer import MinorityCoalescer

import unittest


class MinorityCoalescerTest(unittest.TestCase):
    @property
    def X1(self):
        # Generates an array with categories 3, 4, 5, 6, 7 and occurences of 30%,
        # 30%, 30%, 5% and 5% respectively
        X = np.vstack(
            (
                np.ones((30, 10)) * 3,
                np.ones((30, 10)) * 4,
                np.ones((30, 10)) * 5,
                np.ones((5, 10)) * 6,
                np.ones((5, 10)) * 7,
            )
        )
        for col in range(X.shape[1]):
            np.random.shuffle(X[:, col])
        return X

    @property
    def X2(self):
        # Generates an array with categories 3, 4, 5, 6, 7 and occurences of 5%,
        # 5%, 5%, 35% and 50% respectively
        X = np.vstack(
            (
                np.ones((5, 10)) * 3,
                np.ones((5, 10)) * 4,
                np.ones((5, 10)) * 5,
                np.ones((35, 10)) * 6,
                np.ones((50, 10)) * 7,
            )
        )
        for col in range(X.shape[1]):
            np.random.shuffle(X[:, col])
        return X

    def test_default(self):
        X = self.X1
        X_copy = np.copy(X)
        Y = MinorityCoalescer().fit_transform(X)
        np.testing.assert_array_almost_equal(Y, X_copy)
        # Assert no copies were made
        self.assertEqual(id(X), id(Y))

    def test_coalesce_10_percent(self):
        X = self.X1
        Y = MinorityCoalescer(minimum_fraction=0.1).fit_transform(X)
        for col in range(Y.shape[1]):
            hist = np.histogram(Y[:, col], bins=np.arange(1, 7))
            np.testing.assert_array_almost_equal(hist[0], [10, 0, 30, 30, 30])
        # Assert no copies were made
        self.assertEqual(id(X), id(Y))

    def test_coalesce_10_percent_sparse(self):
        X = scipy.sparse.csc_matrix(self.X1)
        Y = MinorityCoalescer(minimum_fraction=0.1).fit_transform(X)
        # Assert no copies were made
        self.assertEqual(id(X), id(Y))
        Y = Y.todense()
        for col in range(Y.shape[1]):
            hist = np.histogram(Y[:, col], bins=np.arange(1, 7))
            np.testing.assert_array_almost_equal(hist[0], [10, 0, 30, 30, 30])

    def test_invalid_X(self):
        X = self.X1 - 2
        with self.assertRaises(ValueError):
            MinorityCoalescer().fit_transform(X)

    def test_transform_after_fit(self):
        # On both X_fit and X_transf, the categories 3, 4, 5, 6, 7 are present.
        X_fit = self.X1  # Here categories 3, 4, 5 have ocurrence above 10%
        X_transf = self.X2  # Here it is the opposite, just categs 6 and 7 are above 10%

        mc = MinorityCoalescer(minimum_fraction=0.1).fit(X_fit)

        # transform() should coalesce categories as learned during fit.
        # Category distribution in X_transf should be irrelevant.
        Y = mc.transform(X_transf)
        for col in range(Y.shape[1]):
            hist = np.histogram(Y[:, col], bins=np.arange(1, 7))
            np.testing.assert_array_almost_equal(hist[0], [85, 0, 5, 5, 5])



from __future__ import annotations

from typing import Any, Iterable, Mapping, Sequence, Type, cast

import logging.handlers
import multiprocessing
import numbers
import os
import pickle
import time
import traceback
from itertools import accumulate
from pathlib import Path

import numpy as np
import pandas as pd
import pynisher
from sklearn.utils.validation import check_random_state

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.ensemble_building.run import Run, RunID
from autosklearn.ensembles.abstract_ensemble import AbstractEnsemble
from autosklearn.ensembles.ensemble_selection import EnsembleSelection
from autosklearn.metrics import Scorer, calculate_loss, calculate_scores
from autosklearn.util.disk import rmtree
from autosklearn.util.functional import cut, findwhere, roundrobin, split
from autosklearn.util.logging_ import get_named_client_logger
from autosklearn.util.parallel import preload_modules

CANDIDATES_FILENAME = "previous_ensemble_building_candidates.pkl"


class EnsembleBuilder:
    """Builds ensembles out of runs that exist in the Backend

    This is used by EnsembleBuilderManager and created in a dask-client
    every time a run finishes and there is currently no EnsembleBuilder active.
    """

    def __init__(
        self,
        backend: Backend,
        dataset_name: str,
        task_type: int,
        metrics: Sequence[Scorer],
        ensemble_class: Type[AbstractEnsemble] = EnsembleSelection,
        ensemble_kwargs: Mapping[str, Any] | None = None,
        ensemble_nbest: int | float = 50,
        max_models_on_disc: int | float | None = 100,
        seed: int = 1,
        precision: int = 32,
        memory_limit: int | None = 1024,
        read_at_most: int | None = None,
        logger_port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT,
        random_state: int | np.random.RandomState | None = None,
    ):
        """
        Parameters
        ----------
        backend: Backend
            backend to write and read files

        dataset_name: str
            name of dataset

        task_type: int
            type of ML task

        metrics: Sequence[Scorer]
            Metrics to optimize the ensemble for. These must be non-duplicated.

        ensemble_class: Type[AbstractEnsemble]
            Implementation of the ensemble algorithm.

        ensemble_kwargs: Mapping[str, Any] | None
            Arguments passed to the constructor of the ensemble algorithm.

        ensemble_nbest: int | float = 50

            * int: consider only the n best prediction (> 0)

            * float: consider only this fraction of the best, between (0, 1)

            Both with respect to the validation predictions.

        max_models_on_disc: int | float | None = 100
           Defines the maximum number of models that are kept in the disc.
           It defines an upper bound on the models that can be used in the ensemble.

           * int: and dictates the max number of models to keep. (>= 1)

           * float: it will be interpreted as the max megabytes allowed of disc space.
           If the number of ensemble candidates require more disc space than
           this float value, the worst models are deleted to keep within this budget.
           Models and predictions of the worst-performing models will be deleted then.

           * None: the feature is disabled.

        seed: int = 1
            random seed that is used as part of the filename

        precision: int [16 | 32 | 64 | 128] = 32
            precision of floats to read the predictions

        memory_limit: int | None = 1024
            memory limit in mb. If ``None``, no memory limit is enforced.

        read_at_most: int | None = None
            read at most n new prediction files in each iteration. If `None`, will read
            the predictions and calculate losses for all runs that require it.


        logger_port: int = DEFAULT_TCP_LOGGING_PORT
            port that receives logging records

        random_state: int | RandomState | None = None
            An int or RandomState object used for generating the ensemble.
        """
        if isinstance(ensemble_nbest, int) and ensemble_nbest < 1:
            raise ValueError(f"int ensemble_nbest ({ensemble_nbest}) must be (>1)")

        if isinstance(ensemble_nbest, float) and not (0 <= ensemble_nbest <= 1):
            raise ValueError(f"float ensemble_nbest ({ensemble_nbest}) not in (0,1)")

        if max_models_on_disc is not None and max_models_on_disc < 0:
            raise ValueError("max_models_on_disc must be positive or None")

        if read_at_most is not None and (read_at_most < 1 or read_at_most == np.inf):
            raise ValueError("Read at most must be int greater than 1 or None")

        # Setup the logger
        self.logger = get_named_client_logger(name="EnsembleBuilder", port=logger_port)
        self.logger_port = logger_port

        # Log the behaviour
        if ensemble_nbest == 1:
            t = type(ensemble_nbest)
            self.logger.debug(f"Using behaviour when {t} for {ensemble_nbest}:{t}")

        self.seed = seed
        self.metrics = metrics
        self.backend = backend
        self.precision = precision
        self.task_type = task_type
        self.memory_limit = memory_limit
        self.read_at_most = read_at_most
        self.random_state = check_random_state(random_state)
        self.dataset_name = dataset_name
        self.ensemble_class = ensemble_class
        self.ensemble_kwargs = ensemble_kwargs
        self.ensemble_nbest = ensemble_nbest

        # Decide if self.max_models_on_disc is a memory limit or model limit
        self.max_models_on_disc: int | None = None
        self.model_memory_limit: float | None = None

        if isinstance(max_models_on_disc, int):
            self.max_models_on_disc = max_models_on_disc
        elif isinstance(max_models_on_disc, float):
            self.model_memory_limit = max_models_on_disc

        # The starting time of the procedure
        self.start_time: float = 0.0

        # Track the ensemble performance
        self.ensemble_history: list[dict[str, Any]] = []

        # Keep running knowledge of its validation performance
        self.validation_performance_ = np.inf

        # Data we may need
        # TODO: The test data is needlessly loaded but automl_common has no concept of
        # these and is perhaps too rigid
        datamanager: XYDataManager = self.backend.load_datamanager()
        self._X_test: SUPPORTED_FEAT_TYPES | None = datamanager.data.get("X_test", None)
        self._y_test: np.ndarray | None = datamanager.data.get("Y_test", None)
        self._X_ensemble: SUPPORTED_FEAT_TYPES | None = None
        self._y_ensemble: np.ndarray | None = None

    @property
    def previous_candidates_path(self) -> Path:
        """Path to the cached losses we store between runs"""
        return Path(self.backend.internals_directory) / CANDIDATES_FILENAME

    def previous_candidates(self) -> dict[RunID, Run]:
        """Load any previous candidates that were saved from previous runs

        Returns
        -------
        dict[RunID, Run]
            A dictionary from RunId's to the previous candidates
        """
        if self.previous_candidates_path.exists():
            with self.previous_candidates_path.open("rb") as f:
                return pickle.load(f)
        else:
            return {}

    def available_runs(self) -> dict[RunID, Run]:
        """Get a dictionary of all available runs on the filesystem

        Returns
        -------
        dict[RunID, Run]
            A dictionary from RunId's to the available runs
        """
        runs_dir = Path(self.backend.get_runs_directory())
        runs = iter(Run(path=dir) for dir in runs_dir.iterdir() if Run.valid(dir))
        return {run.id: run for run in runs}

    def targets(self, kind: str = "ensemble") -> np.ndarray | None:
        """The ensemble targets used for training the ensemble

        It will attempt to load and cache them in memory but
        return None if it can't.

        Returns
        -------
        np.ndarray | None
            The ensemble targets, if they can be loaded
        """
        if kind == "ensemble":
            if self._y_ensemble is None:
                if os.path.exists(self.backend._get_targets_ensemble_filename()):
                    self._y_ensemble = self.backend.load_targets_ensemble()
            return self._y_ensemble

        elif kind == "test":
            return self._y_test

        else:
            raise NotImplementedError(kind)

    def X_data(self, kind: str = "ensemble") -> SUPPORTED_FEAT_TYPES:
        """The ensemble targets used for training the ensemble

        It will attempt to load and cache them in memory but
        return None if it can't.

        Returns
        -------
        np.ndarray | None
            The ensemble targets, if they can be loaded
        """
        if kind == "ensemble":
            if self._X_ensemble is None:
                if os.path.exists(self.backend._get_input_ensemble_filename()):
                    self._X_ensemble = self.backend.load_input_ensemble()
            return self._X_ensemble

        elif kind == "test":
            return self._X_test

        else:
            raise NotImplementedError(kind)

    def run(
        self,
        iteration: int,
        pynisher_context: str | None = None,
        time_left: float | None = None,
        end_at: float | None = None,
        time_buffer: int = 5,
    ) -> tuple[list[dict[str, Any]], int | float]:
        """Run the ensemble building process

        Parameters
        ----------
        iteration : int
            What iteration to associate with this run

        pynisher_context : str | None = None
            The pynisher context to run in. If None, defaults to
            multiprocessing.get_context(None)

        time_left : float | None = None
            How much time should be left for this run. Either this or `end_at` must
            be provided.

        end_at : float | None = None
            When this run should end. Either this or `time_left` must be provided.

        time_buffer : int = 5
            How much extra time to add as a buffer to this run. This means there is
            always some amount of time to do something useful.

        Returns
        -------
        (ensemble_history, nbest)
        """
        if time_left is None and end_at is None:
            raise ValueError("Must provide either time_left or end_at.")

        elif time_left is not None and end_at is not None:
            raise ValueError("Cannot provide both time_left and end_at.")

        if not self.logger:
            self.logger = get_named_client_logger(
                name="EnsembleBuilder",
                port=self.logger_port,
            )

        process_start_time = time.time()
        while True:

            if time_left is not None:
                time_elapsed = time.time() - process_start_time
                time_left -= time_elapsed
            else:
                assert end_at is not None
                current_time = time.time()
                if current_time > end_at:
                    break
                else:
                    time_left = end_at - current_time

            wall_time_in_s = int(time_left - time_buffer)
            if wall_time_in_s < 1:
                break

            context = multiprocessing.get_context(pynisher_context)
            preload_modules(context)

            safe_ensemble_script = pynisher.enforce_limits(
                wall_time_in_s=wall_time_in_s,
                mem_in_mb=self.memory_limit,
                logger=self.logger,
                context=context,
            )(self.main)

            safe_ensemble_script(time_left, iteration)

            status = safe_ensemble_script.exit_status
            if isinstance(status, pynisher.MemorylimitException):
                # if ensemble script died because of memory error,
                # reduce nbest to reduce memory consumption and try it again

                # ATTENTION: main will start from scratch;
                # all data structures are empty again
                try:
                    self.previous_candidates_path.unlink()
                except:  # noqa E722
                    pass

                if (
                    isinstance(self.ensemble_nbest, numbers.Integral)
                    and self.ensemble_nbest <= 1
                ):
                    if self.read_at_most == 1:
                        self.logger.error(
                            "Memory Exception -- Unable to further reduce the number"
                            " of ensemble members and can no further limit the number"
                            " of ensemble members loaded per iteration, please restart"
                            " Auto-sklearn with a higher value for the argument"
                            f" `memory_limit` (current limit is {self.memory_limit}MB)."
                            " The ensemble builder will keep running to delete files"
                            " from disk in case this was enabled.",
                        )
                        self.ensemble_nbest = 0
                    else:
                        self.read_at_most = 1
                        self.logger.warning(
                            "Memory Exception -- Unable to further reduce the number of"
                            " ensemble members. Now reducing the number of predictions"
                            " per call to read at most to 1."
                        )
                else:
                    if isinstance(self.ensemble_nbest, numbers.Integral):
                        self.ensemble_nbest = max(1, int(self.ensemble_nbest / 2))
                    else:
                        self.ensemble_nbest = self.ensemble_nbest / 2
                    self.logger.warning(
                        "Memory Exception -- restart with "
                        "less ensemble_nbest: %d" % self.ensemble_nbest
                    )
                    return [], self.ensemble_nbest
            elif isinstance(status, pynisher.AnythingException):
                return ([], self.ensemble_nbest)
            else:
                return safe_ensemble_script.result

        return [], self.ensemble_nbest

    def main(
        self,
        time_left: float | None = None,
        iteration: int = 0,
    ) -> tuple[list[dict[str, Any]], int | float]:
        """Run the main loop of ensemble building

        The process is:
        * Load all available runs + previous candidates (if any)
        * Update the loss of those that require
        * From these runs, get a list of candidates
        * Save candidates
        * Delete models that are not candidates
        * Build an ensemble from the candidates if there are new candidates

        Parameters
        ----------
        time_left : float | None = None
            How much time is left for this run

        iteration : int = 0
            The iteration of this run

        Returns
        -------
        (ensemble_history: list[dict[str, Any]], nbest: int | float)
        """
        # Pynisher jobs inside dask 'forget' the logger configuration.
        # So we have to set it up accordingly
        self.logger = get_named_client_logger(
            name="EnsembleBuilder",
            port=self.logger_port,
        )

        if time_left is not None:
            self.start_time = time.time()
            used_time = time.time() - self.start_time
            left_for_iter = time_left - used_time
            itr = iteration if str(iteration) is not None else ""
            self.logger.debug(f"Starting iteration {itr}, time left: {left_for_iter}")

        # Can't load data, exit early
        if not os.path.exists(self.backend._get_targets_ensemble_filename()):
            self.logger.debug(f"No targets for ensemble: {traceback.format_exc()}")
            raise RuntimeError("No targets for ensemble")

        # We will delete runs once we are complete
        deletable_runs: set[Run] = set()

        # Load in information from previous candidates and also runs
        available_runs = self.available_runs()

        # Update runs with information of available previous candidates
        previous_candidates = self.previous_candidates()
        available_runs.update(previous_candidates)

        # We just need the values now, not the key value pairs {run.id: Run}
        runs = list(available_runs.values())

        if len(runs) == 0:
            self.logger.debug("Found no runs")
            raise RuntimeError("Found no runs")

        # We load in `X_data` if we need it
        if any(m._needs_X for m in self.metrics):
            ensemble_X_data = self.X_data("ensemble")

            if ensemble_X_data is None:
                msg = "No `X_data` for 'ensemble' which was required by metrics"
                self.logger.debug(msg)
                raise RuntimeError(msg)
        else:
            ensemble_X_data = None

        # Calculate the loss for those that require it
        requires_update = self.requires_loss_update(runs)
        if self.read_at_most is not None:
            requires_update = requires_update[: self.read_at_most]

        for run in requires_update:
            run.record_modified_times()  # So we don't count as modified next time
            run.losses = {
                metric.name: self.loss(run, metric=metric, X_data=ensemble_X_data)
                for metric in self.metrics
            }

        # Get the dummy and real runs
        dummies, candidates = split(runs, by=lambda r: r.is_dummy())

        # We see if we need to delete any of the real runs before we waste compute
        # on evaluating their candidacy for ensemble building
        if any(candidates):
            candidates, to_delete = self.requires_deletion(
                candidates,
                max_models=self.max_models_on_disc,
                memory_limit=self.model_memory_limit,
            )

            # If there are no candidates left, we just keep the best one, ordered by
            # objectives and finally num_run
            if not any(candidates):
                best = min(
                    to_delete,
                    key=lambda r: (
                        *[r.losses.get(m.name, np.inf) for m in self.metrics],
                        r.num_run,
                    ),
                )
                candidates = [best]
                to_delete.remove(best)

            if any(to_delete):
                self.logger.info(
                    f"Deleting runs {to_delete} due to"
                    f" max_models={self.max_models_on_disc} and/or"
                    f" memory_limit={self.model_memory_limit}"
                )
                deletable_runs.update(to_delete)

        # If there are any candidates, perform candidates selection
        if any(candidates):
            candidates, to_delete = self.candidate_selection(
                runs=candidates,
                dummies=dummies,
                better_than_dummy=True,
                nbest=self.ensemble_nbest,
            )
            if any(to_delete):
                self.logger.info(
                    f"Deleting runs {to_delete} due to"
                    f" nbest={self.ensemble_nbest} or worse than dummies"
                )
                deletable_runs.update(to_delete)
        else:
            candidates = dummies
            self.logger.warning("No runs were available to build an ensemble from")

        # In case we record test predictions and not every model has test predictions,
        # only use the subset of models that has predictions for both the test set and
        # the ensemble optimization set.
        candidates_set = set(candidates)
        test_subset = {r for r in candidates if r.pred_path("test").exists()}

        if len(test_subset) > 0:
            candidates = sorted(test_subset, key=lambda r: r.id)
            test_models = candidates

            to_delete = candidates_set - test_subset
            if any(to_delete):
                self.logger.info(
                    f"Deleting runs {to_delete} due to runs not"
                    ' having "test_predictions" while others do not:'
                    f"\nHave test_predictions = {test_subset}"
                    f"\nNo test_predictions = {to_delete}"
                )
                deletable_runs.update(to_delete)

        else:
            candidates = sorted(candidates_set, key=lambda r: r.id)
            test_models = []

        # Save the candidates for the next round
        with self.previous_candidates_path.open("wb") as f:
            pickle.dump({run.id: run for run in candidates}, f)

        # If there was any change from the previous run, either in terms of
        # runs or one of those runs had its loss updated, then we need to
        # fit the ensemble builder
        previous_candidate_ids = set(previous_candidates)
        current_candidate_ids = set(run.id for run in candidates)
        difference = previous_candidate_ids ^ current_candidate_ids

        was_updated_candidates = list(run in candidates for run in requires_update)

        if not any(difference) and not any(was_updated_candidates):
            self.logger.info("All ensemble candidates the same, no update required")
            return self.ensemble_history, self.ensemble_nbest

        targets = cast(np.ndarray, self.targets("ensemble"))  # Sure they exist

        ensemble = self.fit_ensemble(
            candidates=candidates,
            targets=targets,
            runs=runs,
            ensemble_class=self.ensemble_class,
            ensemble_kwargs=self.ensemble_kwargs,
            X_data=ensemble_X_data,
            task=self.task_type,
            metrics=self.metrics,
            precision=self.precision,
            random_state=self.random_state,
        )

        self.logger.info(str(ensemble))
        ens_perf = ensemble.get_validation_performance()
        self.validation_performance_ = min(self.validation_performance_, ens_perf)
        self.backend.save_ensemble(
            ensemble=ensemble, idx=iteration, seed=self.seed  # type: ignore
        )

        performance_stamp = {"Timestamp": pd.Timestamp.now()}

        for kind, score_name, models in [
            ("ensemble", "optimization", candidates),
            ("test", "test", test_models),
        ]:
            if len(models) == 0:
                continue

            pred_targets = self.targets(kind)
            if pred_targets is None:
                self.logger.warning(f"No ensemble targets for {kind}")
                continue

            run_preds = [r.predictions(kind, precision=self.precision) for r in models]
            pred = ensemble.predict(run_preds)

            if any(m._needs_X for m in self.metrics):
                X_data = self.X_data(kind)
                if X_data is None:
                    msg = f"No `X` data for '{kind}' which was required by metrics"
                    self.logger.debug(msg)
                    raise RuntimeError(msg)
            else:
                X_data = None

            scores = calculate_scores(
                solution=pred_targets,
                prediction=pred,
                task_type=self.task_type,
                metrics=self.metrics,
                X_data=X_data,
                scoring_functions=None,
            )

            # TODO only one metric in history
            #
            #   We should probably return for all metrics but this makes
            #   automl::performance_history a lot more complicated, will
            #   tackle in a future PR
            first_metric = self.metrics[0]
            performance_stamp[f"ensemble_{score_name}_score"] = scores[
                first_metric.name
            ]

        # Add the performance stamp to the history
        self.ensemble_history.append(performance_stamp)

        # Lastly, delete any runs that need to be deleted. We save this as the last step
        # so that we have an ensemble saved that is up to date. If we do not do so,
        # there could be runs deleted that are in th previous ensemble and we do not
        # manage to update the ensemble due to a crash or the process being killed
        # before it could be updated
        self.delete_runs(deletable_runs)

        return self.ensemble_history, self.ensemble_nbest

    def requires_loss_update(
        self,
        runs: Sequence[Run],
        metrics: Sequence[Scorer] | None = None,
    ) -> list[Run]:
        """

        Parameters
        ----------
        runs : Sequence[Run]
            The runs to process

        metrics: Sequence[Scorer] | None = None
            The metrics to check for, defaults to builders

        Returns
        -------
        list[Run]
            The runs that require a loss to be calculated
        """
        metrics = metrics if metrics is not None else self.metrics

        queue = []
        for run in sorted(runs, key=lambda run: run.recorded_mtimes["ensemble"]):
            if any(metric.name not in run.losses for metric in metrics):
                queue.append(run)

            elif run.was_modified():
                self.logger.debug(f"{run.id} had its predictions modified?")
                queue.append(run)

        return queue

    def candidate_selection(
        self,
        runs: Sequence[Run],
        dummies: Run | list[Run],
        *,
        better_than_dummy: bool = False,
        nbest: int | float | None = None,
        metrics: Sequence[Scorer] | None = None,
    ) -> tuple[list[Run], set[Run]]:
        """Get a list of candidates from `runs`, garuanteeing at least one

        Applies a set of reductions in order of parameters to reach a set of final
        candidates.

        Expects at least one `dummies` run.

        Parameters
        ----------
        runs : Sequence[Run]
            The runs to evaluate candidates from.

        dummies: Run | Sequence[Run]
            The dummy run to base from

        better_than_dummy: bool = False
            Whether the run must be better than the best dummy run to be a candidate.
            In the case where there are no candidates left, the dummies will then be
            used.

        nbest : int | float | None
            The nbest models to select. If `int`, acts as an absolute limit.
            If `float`, acts as a percentage of available candidates.

        metrics : Sequence[Scorer] | None = None
            The metrics to consider, defaults to the builders

        Returns
        -------
        (candidates: list[Run], discarded: set[Run])
            A tuple of runs that are candidates and also those that didn't make it
        """
        metrics = metrics if metrics else self.metrics

        if isinstance(dummies, Run):
            dummies = [dummies]

        assert len(dummies) > 0 and len(runs) > 0, "At least 1 real run and dummy run"

        all_discarded: set[Run] = set()

        # We filter out all runs that don't have any predictions for the ensemble
        candidates, discarded = split(
            runs,
            by=lambda run: run.pred_path("ensemble").exists(),
        )
        all_discarded.update(discarded)

        if len(candidates) == 0:
            self.logger.debug("No runs with predictions on ensemble set, using dummies")
            return dummies, all_discarded

        for run in discarded:
            self.logger.warning(f"Have no ensemble predictions for {run}")

        # Get all the ones that have all metrics and tangible losses
        has_metrics = lambda r: all(m.name in r.losses for m in metrics)
        tangible_losses = lambda r: all(loss < np.inf for loss in r.losses.values())

        candidates, discarded = split(
            candidates,
            by=lambda r: has_metrics(r) and tangible_losses(r),
        )
        all_discarded.update(discarded)

        if len(candidates) == 0:
            self.logger.debug("No runs with a usable loss, using dummies")
            return dummies, all_discarded

        if better_than_dummy:

            # In the single objective case, they must explicitly be better than dummy
            if len(metrics) == 1:
                metric = metrics[0]

                best_dummy = min(dummies, key=lambda d: d.losses[metric.name])
                self.logger.debug(f"Using dummy {best_dummy} to filter candidates")

                candidates, discarded = split(
                    candidates,
                    by=lambda r: r.losses[metric.name] < best_dummy.losses[metric.name],
                )

            # In the multiobjective case, they must be better than at least one of the
            # best dummies, where there is a best dummy for each metric
            else:
                best_dummies = {
                    metric.name: min(dummies, key=lambda d: d.losses[metric.name])
                    for metric in metrics
                }
                self.logger.debug(f"Using dummies {best_dummies} to filter candidates")

                candidates, discarded = split(
                    candidates,
                    by=lambda r: any(
                        r.losses[metric_name] < best_dummy.losses[metric_name]
                        for metric_name, best_dummy in best_dummies.items()
                    ),
                )

            all_discarded.update(discarded)

            # If there are no real candidates left, use the dummies
            if len(candidates) == 0:
                self.logger.warning(
                    "No models better than random - using Dummy losses!"
                    f"\n\tModels besides current dummy model: {len(candidates)}"
                    f"\n\tDummy models: {len(dummies)}",
                )
                return dummies, all_discarded

        if nbest is not None:
            # We order the runs according to a roundrobin of the metrics
            # i.e. the 1st run will be best in metric[0],
            #      the 2nd run will be best in metric[1],
            #      the 3rd run will be second best in metric[0],
            #      the 4th run will be second best in metric[1],
            #       ... whoops, what would be the 5th run, third best in metric[0] was
            #           also the same as the 4th run. This was skipped
            #      the 5th run will be third best in metric[1]  **as we skipped above**
            #
            # With removing duplicates, this means if a run was already contained, it
            # will skip to the next member in the roundrobin fashion.
            sorted_candidates = [
                sorted(
                    candidates,
                    key=lambda r: (r.losses.get(m.name, np.inf), r.num_run),
                )
                for m in metrics
            ]
            candidates = list(roundrobin(*sorted_candidates, duplicates=False))

            # Determine how many to keep, always keeping one
            if isinstance(nbest, float):
                nkeep = int(len(candidates) * nbest)
            else:
                nkeep = nbest

            candidates, discarded = cut(candidates, nkeep)
            self.logger.info(f"Discarding {len(discarded)}/{len(candidates)} runs")

            # Always preserve at least one, the best
            if len(candidates) == 0:
                candidates, discared = cut(discarded, 1)
                self.logger.warning("nbest too aggresive, using single best")

            all_discarded.update(discarded)

        return candidates, all_discarded

    def fit_ensemble(
        self,
        candidates: Sequence[Run],
        runs: Sequence[Run],
        *,
        targets: np.ndarray | None = None,
        ensemble_class: Type[AbstractEnsemble] = EnsembleSelection,
        ensemble_kwargs: Mapping[str, Any] | None = None,
        X_data: SUPPORTED_FEAT_TYPES | None = None,
        task: int | None = None,
        metrics: Sequence[Scorer] | None = None,
        precision: int | None = None,
        random_state: int | np.random.RandomState | None = None,
    ) -> AbstractEnsemble:
        """Fit an ensemble from the provided runs.

        Note
        ----
        Expects all runs to have the "ensemble" predictions present

        Parameters
        ----------
        candidates: Sequence[Run]
            List of runs to build an ensemble from

        runs: Sequence[Run]
            List of all runs (also pruned ones and dummy runs)

        targets: np.ndarray | None = None
            The targets to build the ensemble with

        ensemble_class: Type[AbstractEnsemble]
            Implementation of the ensemble algorithm.

        ensemble_kwargs: Mapping[str, Any] | None
            Arguments passed to the constructor of the ensemble algorithm.

        X_data: SUPPORTED_FEAT_TYPES | None = None
            The base level data.

        task: int | None = None
            The kind of task performed

        metrics: Sequence[Scorer] | None = None
            Metrics to optimize the ensemble for.

        precision: int | None = None
            The precision with which to load run predictions

        random_state: int | RandomState | None = None
            The random state to use

        Returns
        -------
        AbstractEnsemble
        """
        # Validate we have targets if None specified
        if targets is None:
            targets = self.targets("ensemble")
            if targets is None:
                path = self.backend._get_targets_ensemble_filename()
                raise ValueError(f"`fit_ensemble` could not find any targets at {path}")

        ensemble_class = (
            ensemble_class if ensemble_class is not None else self.ensemble_class
        )

        # Create the ensemble_kwargs, favouring in order:
        # 1) function kwargs, 2) function params 3) init_kwargs 4) init_params

        # Collect func params in dict if they're not None
        params = {
            k: v
            for k, v in [
                ("task_type", task),
                ("metrics", metrics),
                ("random_state", random_state),
            ]
            if v is not None
        }

        kwargs = {
            "backend": self.backend,
            "task_type": self.task_type,
            "metrics": self.metrics,
            "random_state": self.random_state,
            **(self.ensemble_kwargs or {}),
            **params,
            **(ensemble_kwargs or {}),
        }

        ensemble = ensemble_class(**kwargs)  # type: AbstractEnsemble

        self.logger.debug(f"Fitting ensemble on {len(candidates)} models")
        start_time = time.time()

        precision = precision if precision is not None else self.precision
        predictions_train = [
            run.predictions("ensemble", precision=precision) for run in candidates
        ]

        ensemble.fit(
            base_models_predictions=predictions_train,
            X_data=X_data,
            true_targets=targets,
            model_identifiers=[run.id for run in candidates],
            runs=runs,
        )

        duration = time.time() - start_time
        self.logger.debug(f"Fitting the ensemble took {duration} seconds.")
        return ensemble

    def requires_deletion(
        self,
        runs: Sequence[Run],
        *,
        max_models: int | None = None,
        memory_limit: float | None = None,
        metrics: Sequence[Scorer] | None = None,
    ) -> tuple[list[Run], set[Run]]:
        """Cut a list of runs into those to keep and those to delete

        If neither params are specified, this method should do nothing.

        In the case of multiple metrics, we interleave these according to the order of
        the provided metrics.

        Parameters
        ----------
        runs : Sequence[Run]
            The runs to check

        max_models : int | None = None
            The maximum amount of models to have on disk. Leave `None` for no effect

        memory_limit : float | None = None
            The memory limit in MB, leave `None` for no effect

        metrics: Sequence[Scorer] | None = None
            The metrics to consider for the runs, defaults to `self.metrics` passed
            during construction if `None`.

        Returns
        -------
        (keep: list[Run], delete: set[Run])
            The list of runs to keep and those to delete
        """
        if memory_limit is None and max_models is None:
            return list(runs), set()

        metrics = metrics if metrics is not None else self.metrics

        # We order the runs according to a roundrobin of the metrics
        # i.e. the 1st run will be best in metric[0],
        #      the 2nd run will be best in metric[1],
        #      the 3rd run will be second best in metric[0],
        #      the 4th run will be second best in metric[1],
        #       ... whoops, what would be the 5th run, third best in metric[0] was also
        #       ... the same as the 4th run. This was skipped
        #      the 5th run will be third best in metric[1]  **as we skipped above**
        #
        # With removing duplicates, this means if a run was already contained, it will
        # skip to the next member in the roundrobin fashion.
        sorted_runs = [
            sorted(runs, key=lambda r: (r.losses.get(metric.name, np.inf), r.num_run))
            for metric in metrics
        ]
        keep = list(roundrobin(*sorted_runs, duplicates=False))

        delete: set[Run] = set()

        if max_models is not None and max_models < len(runs):
            keep, to_delete = cut(keep, max_models)

            if any(to_delete):
                self.logger.info(
                    "Limiting num of models via `max_models_on_disc`"
                    f" max_models={max_models}"
                    f" remaining={len(keep)}"
                    f" discarded={len(to_delete)}"
                )
                delete.update(to_delete)

        if memory_limit is not None:
            largest = max(runs, key=lambda r: r.mem_usage)
            cutoff = memory_limit - largest.mem_usage

            accumulated_mem_usage = accumulate(r.mem_usage for r in runs)

            cutpoint = findwhere(accumulated_mem_usage, lambda mem: mem > cutoff)
            keep, to_delete = cut(keep, cutpoint)

            if any(to_delete):
                self.logger.info(
                    "Limiting num of models via `memory_limit`"
                    f" memory_limit={memory_limit}"
                    f" cutoff={cutoff}"
                    f" largest={largest.mem_usage}"
                    f" remaining={len(keep)}"
                    f" discarded={len(to_delete)}"
                )
                delete.update(to_delete)

        return keep, delete

    def loss(
        self,
        run: Run,
        metric: Scorer,
        *,
        X_data: SUPPORTED_FEAT_TYPES | None = None,
        kind: str = "ensemble",
    ) -> float:
        """Calculate the loss for a run

        Parameters
        ----------
        run: Run
            The run to calculate the loss for

        metric: Scorer
            The metric to calculate the loss of

        X_data: SUPPORTED_FEAT_TYPES | None = None
            Any X_data required to be passed to the metric

        kind: str = "ensemble"
            The kind of targets to use for the run

        Returns
        -------
        float
            The loss for the run
        """
        targets = self.targets(kind)
        if targets is None:
            self.logger.error(f"No targets of {kind}")
            return np.inf

        try:
            predictions = run.predictions(kind, precision=self.precision)
            loss: float = calculate_loss(
                solution=targets,
                prediction=predictions,
                task_type=self.task_type,
                metric=metric,
                X_data=X_data,
            )
        except Exception as e:
            tb = traceback.format_exc()
            self.logger.error(f"Error getting loss `{metric.name}` for {run}:{e}{tb}")
            loss = np.inf
        finally:
            return loss

    def delete_runs(self, runs: Iterable[Run]) -> None:
        """Delete runs

        Will not delete dummy runs

        Parameters
        ----------
        runs : Sequence[Run]
            The runs to delete
        """
        items = iter(run for run in runs if not run.is_dummy() and run.dir.exists())
        for run in items:
            try:
                rmtree(run.dir, atomic=True)
                self.logger.info(f"Deleted files for {run}")
            except Exception as e:
                self.logger.error(f"Failed to delete files for {run}: \n{e}")



from typing import Dict, Optional, Tuple, Union

import numpy as np
import sklearn.feature_selection
from ConfigSpace.configuration_space import ConfigurationSpace

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class VarianceThreshold(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self, random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        # VarianceThreshold does not support fit_transform (as of 0.19.1)!
        self.random_state = random_state

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "VarianceThreshold":
        self.preprocessor = sklearn.feature_selection.VarianceThreshold(threshold=0.0)
        self.preprocessor = self.preprocessor.fit(X)
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "Variance Threshold",
            "name": "Variance Threshold (constant feature removal)",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        return cs



from __future__ import annotations

from typing import Callable, Hashable, Iterable, Iterator, TypeVar

from functools import reduce
from itertools import chain, cycle, islice, tee

T = TypeVar("T")


def intersection(*items: Iterable[T]) -> set[T]:
    """Does an intersection over all collection of items

    ..code:: python

        ans = intersection(["a", "b", "c"], "ab", ("b", "c"))

        items = [(1, 2, 3), (2, 3), (4, 5)]
        ans = intesection(*items)

    Parameters
    ----------
    *items : Iterable[T]
        A list of lists

    Returns
    -------
    Set[T]
        The intersection of all items
    """
    if len(items) == 0:
        return set()

    return set(reduce(lambda s1, s2: set(s1) & set(s2), items, items[0]))


def cut(
    itr: Iterable[T],
    where: int | Callable[[T], bool],
) -> tuple[list[T], list[T]]:
    """Cut a list in two at a given index or predicate

    Parameters
    ----------
    lst : Iterable[T]
        An iterable of items

    at : int | Callable[[T], bool]
        Where to split at, either an index or a predicate

    Returns
    -------
    tuple[list[T], list[T]]
        The split items
    """
    if isinstance(where, int):
        lst = list(itr)
        return lst[:where], lst[where:]

    a = []
    itr2 = iter(itr)
    broke = False
    for x in itr2:
        if not where(x):
            a.append(x)
        else:
            broke = True
            break

    if broke:
        return a, [x] + list(itr2)
    else:
        return a, []


def split(
    lst: Iterable[T],
    by: Callable[[T], bool],
) -> tuple[list[T], list[T]]:
    """Split a list in two based on a predicate.

    Note
    ----
    First element can not contain None

    Parameters
    ----------
    lst : Iterable[T]
        The iterator to split

    by : Callable[[T], bool]
        The predicate to split it on

    Returns
    -------
    (a: list[T], b: list[T])
        a is where the func is True and b is where the func was False.
    """
    a = []
    b = []
    for x in lst:
        if by(x):
            a.append(x)
        else:
            b.append(x)

    return a, b


def bound(val: float, bounds: tuple[float, float]) -> float:
    """Bounds a value between a low and high

    .. code:: python

        x = bound(14, low=0, high=13.1)
        # x == 13.1

    Parameters
    ----------
    val : float
        The value to bound

    bounds: tuple[foat, float]
        The bounds to bound the value between (low, high)

    Returns
    -------
    float
        The bounded value
    """
    return max(bounds[0], min(val, bounds[1]))


def findwhere(itr: Iterable[T], func: Callable[[T], bool], *, default: int = -1) -> int:
    """Find the index of the next occurence where func is True.

    Parameters
    ----------
    itr : Iterable[T]
        The iterable to search over

    func : Callable[[T], bool]
        The function to use

    default : int = -1
        The default value to give if no value was found where func was True

    Returns
    -------
    int
        The first index where func was True
    """
    return next((i for i, t in enumerate(itr) if func(t)), default)


def pairs(itr: Iterable[T]) -> Iterator[tuple[T, T]]:
    """An iterator over pairs of items in the iterator

    ..code:: python

        # Check if sorted
        if all(a < b for a, b in pairs(items)):
            ...

    Parameters
    ----------
    itr : Iterable[T]
        An itr of items

    Returns
    -------
    Iterable[tuple[T, T]]
        An itr of sequential pairs of the items
    """
    itr1, itr2 = tee(itr)

    # Skip first item
    _ = next(itr2)

    # Check there is a second element
    peek = next(itr2, None)
    if peek is None:
        raise ValueError("Can't create a pair from iterable with 1 item")

    # Put it back in
    itr2 = chain([peek], itr2)

    return iter((a, b) for a, b in zip(itr1, itr2))


def roundrobin(
    *iterables: Iterable[T],
    duplicates: bool = True,
    key: Callable[[T], Hashable] | None = None,
) -> Iterator[T]:
    """Performs a round robin iteration of several iterables

    Adapted from https://docs.python.org/3/library/itertools.html#recipes

    ..code:: python

        colours = ["orange", "red", "green"]
        fruits = ["apple", "banana", "orange"]

        list(roundrobin(colors, fruits))
        # ["orange", "apple", "red", "banana", "green", "orange"]

        list(roundrobin(colors, fruits, duplicates=False))
        # ["orange", "apple", "red", "banana", "green"]

    Parameters
    ----------
    *iterables: Iterable[T]
        Any amount of iterables

    duplicates: bool = True
        Whether duplicates are allowed

    key: Callable[[T], Hashable] | None = None
        A key to use when checking for duplicates

    Returns
    -------
    Iterator[T]
        A round robin iterator over the iterables passed
    """
    active_iterators = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)  # roundrobin workhorse

    # We split up the algorithm into removing duplicates and not. This removes the if
    # statement from within the loop at the cost of duplication.
    if duplicates:
        while active_iterators > 0:
            try:
                for nxt in nexts:
                    yield nxt()
            except StopIteration:
                active_iterators -= 1
                nexts = cycle(islice(nexts, active_iterators))

    else:
        seen = set()
        key = key if key is not None else lambda x: x  # Identity if None

        while active_iterators > 0:
            try:
                for nxt in nexts:
                    item = nxt()
                    identifier = key(item)

                    if identifier not in seen:
                        seen.add(identifier)
                        yield item

            except StopIteration:
                active_iterators -= 1
                nexts = cycle(islice(nexts, active_iterators))



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    import numpy as np
    from smac.optimizer.smbo import SMBO
    from smac.runhistory.runhistory import RunInfo, RunValue

    from autosklearn.automl import AutoMLClassifier


def test_early_stopping(
    make_automl_classifier: Callable[..., AutoMLClassifier],
    make_sklearn_dataset: Callable[..., tuple[np.ndarray, ...]],
) -> None:
    """
    Expects
    -------
    * Should early after fitting 2 models
    """

    def callback(
        smbo: SMBO,
        run_info: RunInfo,
        result: RunValue,
        time_left: float,
    ) -> bool | None:
        if int(result.additional_info["num_run"]) >= 2:
            return False

    automl = make_automl_classifier(get_trials_callback=callback)

    X_train, Y_train, X_test, Y_test = make_sklearn_dataset("iris")
    automl.fit(X_train, Y_train)

    assert len(automl.runhistory_.data) == 2



import numpy as np
from scipy import sparse

from autosklearn.pipeline.components.data_preprocessing.feature_type_categorical import (  # noqa: E501
    CategoricalPreprocessingPipeline,
)

import pytest
import unittest


class CategoricalPreprocessingPipelineTest(unittest.TestCase):
    def test_data_type_consistency(self):
        X = np.random.randint(3, 6, (3, 4))
        Y = CategoricalPreprocessingPipeline().fit_transform(X)
        self.assertFalse(sparse.issparse(Y))

        X = sparse.csc_matrix(
            ([3, 6, 4, 5], ([0, 1, 2, 1], [3, 2, 1, 0])), shape=(3, 4)
        )
        Y = CategoricalPreprocessingPipeline().fit_transform(X)
        self.assertTrue(sparse.issparse(Y))

    def test_fit_transform(self):
        X = np.array([[1, 2, 1], [3, 1, 1], [2, 9, np.nan]])
        Y = np.array(
            [
                [1, 0, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 1, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 1, 1, 0],
            ]
        )
        # dense input
        # Notice the X.copy() here as the imputation
        # is in place to save resources
        Yt = CategoricalPreprocessingPipeline().fit_transform(X.copy())
        np.testing.assert_array_equal(Yt, Y)
        # sparse input
        X_sparse = sparse.csc_matrix(X)
        Yt = CategoricalPreprocessingPipeline().fit_transform(X_sparse)
        Yt = Yt.todense()
        np.testing.assert_array_equal(Yt, Y)

    def test_transform(self):
        X1 = np.array([[1, 2, 0], [3, 0, 0], [2, 9, np.nan]])
        Y1 = np.array(
            [
                [1, 0, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 1, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 1, 1, 0],
            ]
        )
        X2 = np.array([[2, 2, 1], [3, 0, 0], [2, np.nan, np.nan]])
        Y2 = np.array(
            [
                [0, 1, 0, 0, 1, 0, 0, 0],
                [0, 0, 1, 1, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 0, 1, 0],
            ]
        )
        X3 = np.array([[3, np.nan, 0], [3, 9, np.nan], [2, 2, 5]])
        Y3 = np.array(
            [
                [0, 0, 1, 0, 0, 0, 0, 1],
                [0, 0, 1, 0, 0, 1, 1, 0],
                [0, 1, 0, 0, 1, 0, 0, 0],
            ]
        )
        # "fit"
        CPPL = CategoricalPreprocessingPipeline()
        CPPL.fit_transform(X1)
        # Transform what was fitted
        Y1t = CPPL.transform(X1)
        np.testing.assert_array_equal(Y1t, Y1)
        # Transform a new dataset with categories not seen during fit
        Y2t = CPPL.transform(X2)
        np.testing.assert_array_equal(Y2t, Y2)
        # And again with yet a different dataset
        Y3t = CPPL.transform(X3)
        np.testing.assert_array_equal(Y3t, Y3)

    def test_transform_with_coalescence(self):
        # Generates an array with categories 0, 20, 5, 6, 10, and occurences of 60%,
        # 30%, 19% 0.5% and 0.5% respectively
        X = np.vstack(
            (
                np.ones((120, 10)) * 0,
                np.ones((60, 10)) * 20,
                np.ones((18, 10)) * 5,
                np.ones((1, 10)) * 6,
                np.ones((1, 10)) * 10,
            )
        )
        for col in range(X.shape[1]):
            np.random.shuffle(X[:, col])

        CPPL = CategoricalPreprocessingPipeline()
        Y1t = CPPL.fit_transform(X)
        # From the 5 original categories, 2 are coalesced, remaining 4.
        # Dataset has 10 cols, therefore Y must have 40 (i.e. 4 x 10) cols
        self.assertEqual(Y1t.shape, (200, 40))
        # Consistency check:
        Y2t = CPPL.transform(X)
        np.testing.assert_array_equal(Y1t, Y2t)

    @pytest.mark.xfail(
        reason=(
            "Encoding step does not support sparse matrices to convert negative labels"
            " to positive ones as it does with non-sparse matrices"
        )
    )
    def test_transform_with_sparse_column_with_negative_labels(self):
        X = sparse.csr_matrix([[0], [-1]])
        CategoricalPreprocessingPipeline().fit_transform(X)



from typing import List, Optional, Type, Union, cast

import logging
import warnings

import numpy as np
import pandas as pd
import sklearn.utils
from pandas.api.types import is_numeric_dtype
from pandas.core.dtypes.base import ExtensionDtype
from scipy.sparse import spmatrix
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import OrdinalEncoder
from sklearn.utils.multiclass import type_of_target

from autosklearn.util.logging_ import PickableLoggerAdapter

SUPPORTED_TARGET_TYPES = Union[List, pd.Series, pd.DataFrame, np.ndarray, spmatrix]
SUPPORTED_XDATA_TYPES = Union[pd.Series, pd.DataFrame, np.ndarray, spmatrix]


class TargetValidator(BaseEstimator):
    """A class to pre-process targets.

    It validates the data provided during fit to make sure it matches Sklearn
    expectation as well as encoding the targets in case of classification

    Attributes
    ----------
    is_classification: bool
        A bool that indicates if the validator should operate in classification mode.
        During classification, the targets are encoded.

    encoder: Optional[BaseEstimator]
        Host a encoder object if the data requires transformation (for example, if
        provided a categorical column in a pandas DataFrame).

    enc_columns: List[str]
        List of columns that where encoded
    """

    def __init__(
        self,
        is_classification: bool = False,
        logger: Optional[PickableLoggerAdapter] = None,
    ) -> None:
        self.is_classification = is_classification

        self.data_type = None  # type: Optional[type]

        # NOTE: type update
        #   Encoders don't seems to have a unified base class for
        #   the methods 'transform'. Could make a `prototype` class that
        #   duck types for the 'transform', 'fit_transform' methods
        #   This is however fine while we only use ordinal encoder
        self.encoder = None  # type: Optional[OrdinalEncoder]

        self.out_dimensionality = None  # type: Optional[int]
        self.type_of_target = None  # type: Optional[str]

        self.logger = logger if logger is not None else logging.getLogger(__name__)

        # Store the dtype for remapping to correct type
        self.dtype = None  # type: Optional[Type]

        self._is_fitted = False

    def fit(
        self,
        y_train: Union[List, np.ndarray, pd.Series, pd.DataFrame],
        y_test: Optional[Union[List, np.ndarray, pd.Series, pd.DataFrame]] = None,
    ) -> "TargetValidator":
        """
        Validates and fit a categorical encoder (if needed) to the targets
        The supported data types are List, numpy arrays and pandas DataFrames.

        Parameters
        ----------
            y_train: SUPPORTED_TARGET_TYPES
                A set of targets set aside for training
            y_test: Union[SUPPORTED_TARGET_TYPES]
                A hold out set of data used of the targets. It is also used to fit the
                categories of the encoder.
        """
        # Check that the data is valid
        self._check_data(y_train)

        shape = np.shape(y_train)
        if y_test is not None:
            self._check_data(y_test)

            if len(shape) != len(np.shape(y_test)) or (
                len(shape) > 1 and (shape[1] != np.shape(y_test)[1])
            ):
                raise ValueError(
                    "The dimensionality of the train and test"
                    " targets do not match"
                    f" train {np.shape(y_train)}"
                    f" != test {np.shape(y_test)}"
                )

            if isinstance(y_train, pd.DataFrame):
                if not isinstance(y_test, pd.DataFrame):
                    y_test = pd.DataFrame(y_test)

                if y_train.columns.tolist() != y_test.columns.tolist():
                    raise ValueError(
                        "Train and test targets must both have the"
                        f" same columns, yet y={y_train.columns}"
                        f" and y_test={y_test.columns}"
                    )

                if list(y_train.dtypes) != list(y_test.dtypes):
                    raise ValueError(
                        "Train and test targets must both have the same dtypes"
                    )

        if self.out_dimensionality is None:
            self.out_dimensionality = 1 if len(shape) == 1 else shape[1]
        else:
            _n_outputs = 1 if len(shape) == 1 else shape[1]
            if self.out_dimensionality != _n_outputs:
                raise ValueError(
                    "Number of outputs changed from %d to %d!"
                    % (self.out_dimensionality, _n_outputs)
                )

        # Fit on the training data
        self._fit(y_train, y_test)

        self._is_fitted = True

        return self

    def _fit(
        self,
        y_train: Union[List, np.ndarray, pd.Series, pd.DataFrame],
        y_test: Optional[Union[List, np.ndarray, pd.Series, pd.DataFrame]] = None,
    ) -> "TargetValidator":
        """
        If dealing with classification, this utility encodes the targets.

        It does so by also using the classes from the test data, to prevent encoding
        errors

        Parameters
        ----------
            y_train: SUPPORTED_TARGET_TYPES
                The labels of the current task. They are going to be encoded in case
                of classification
            y_test: Optional[SUPPORTED_TARGET_TYPES]
                A holdout set of labels
        """
        if not self.is_classification or self.type_of_target == "multilabel-indicator":
            # Only fit an encoder for classification tasks
            # Also, encoding multilabel indicator data makes the data multiclass
            # Let the user employ a MultiLabelBinarizer if needed
            self.encoder = None
            return self

        # Validate the shape of the input
        shape = np.shape(y_train)
        ndim = len(shape)
        if ndim > 1 and shape[1] != 1:
            # We should not reach this if statement, we check for type of targets before
            raise ValueError(
                "Multi-dimensional classification is not yet"
                " supported. Encoding multidimensional data"
                " converts multiple columns to a 1 dimensional encoding."
                f" Data involved = {shape}/{self.type_of_target}"
            )

        # Create the encoder
        self.encoder = OrdinalEncoder(handle_unknown="error")

        # Clear typing to just numpy arrays and pandas
        if isinstance(y_train, List):
            y = np.asarray(y_train)
        else:
            y = y_train

        # Attempt to store dtype if it's numeric, we use this in
        # inverse_transform to try corretly restore it's dtype
        if isinstance(y, pd.Series):
            if isinstance(y.dtype, ExtensionDtype):
                warnings.warn(
                    "Fitting transformer with a pandas series which"
                    f" has the dtype {y.dtype}. Inverse transform"
                    " may not be able preserve dtype when converting"
                    " to np.ndarray"
                )
            if is_numeric_dtype(y.dtype):
                self.dtype = y.dtype
        elif isinstance(y, pd.DataFrame):
            if is_numeric_dtype(y.dtypes[0]):
                self.dtype = y.dtypes[0]
        else:
            if is_numeric_dtype(y.dtype):
                self.dtype = y.dtype

        # Merge y_test and y_train for encoding
        if y_test is not None:
            if isinstance(y, (pd.Series, pd.DataFrame)):
                if isinstance(y, pd.Series):
                    y_test = pd.Series(y_test)
                else:
                    y_test = pd.DataFrame(y_test)

                y = pd.concat([y, y_test], ignore_index=True, sort=False)
            else:
                y = np.concatenate([y, y_test])

        # If one dimensional, we need to reshape it
        # [1, 2, 3] -> [[1], [2], [3]]
        if y.ndim == 1:
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.to_numpy()

            y = y.reshape(-1, 1)

        # Fit encoder
        self.encoder.fit(y)

        return self

    def transform(
        self,
        y: Union[List, pd.Series, pd.DataFrame, np.ndarray],
    ) -> np.ndarray:
        """
        Validates and fit a categorical encoder (if needed) to the features.
        The supported data types are List, numpy arrays and pandas DataFrames.

        Parameters
        ----------
            y: SUPPORTED_TARGET_TYPES
                A set of targets that are going to be encoded if the current task
                is classification
        Return
        ------
            np.ndarray:
                The transformed array
        """
        if not self._is_fitted:
            raise NotFittedError("TargetValidator must have fit() called first")

        # Check the data here so we catch problems on new test data
        self._check_data(y)

        # Clear the types List and DataFrame off of y
        if isinstance(y, List):
            y_transformed = np.asarray(y)
        elif isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y_transformed = y.to_numpy()
        else:
            y_transformed = y

        if self.encoder is None:
            return y_transformed

        # The Ordinal encoder expects a 2 dimensional input.
        # The targets are 1 dimensional, so reshape to match the expected shape
        if y_transformed.ndim == 1:
            y_transformed = y_transformed.reshape(-1, 1)
            y_transformed = self.encoder.transform(y_transformed).reshape(-1)
        else:
            y_transformed = self.encoder.transform(y_transformed)

        # check_array ensures correct numerical features for array
        y_transformed = sklearn.utils.check_array(
            y_transformed,
            force_all_finite=True,
            accept_sparse="csr",
            ensure_2d=False,
        )

        # Ensure we flatten any arrays [[1], [2], [1], [1], ...] to [1,2,1,1,...]
        if y_transformed.ndim == 2 and y_transformed.shape[1] == 1:
            y_transformed = y_transformed.ravel()

        return y_transformed

    def inverse_transform(
        self,
        y: Union[List, pd.Series, pd.DataFrame, np.ndarray],
    ) -> np.ndarray:
        """Revert any encoding transformation done on a target array.

        Parameters
        ----------
        y: Union[List, np.ndarray, pd.DataFrame, pd.Series]
            Target array to be transformed back to original form before encoding

        Return
        ------
        np.ndarray:
            The transformed array
        """
        if not self._is_fitted:
            raise NotFittedError("TargetValidator must have fit() called first")

        # If there's no encoder, just return it as the expected ndarray
        if self.encoder is None:
            return np.asarray(y)

        # If it's one dimensional, we need to convert it
        # [1,2,3] -> [[1], [2], [3]]
        shape = np.shape(y)
        if len(shape) == 1:

            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.to_numpy()
            elif isinstance(y, List):
                y = np.asarray(y)

            # Perform the transformation and reshape it
            y = y.reshape(-1, 1)
            y_inv = self.encoder.inverse_transform(y)
            y_inv = y_inv.reshape(-1)

        else:
            # Otherwise we can can just invert it as expected
            y_inv = self.encoder.inverse_transform(y)

        # Inverse transform returns a numpy array of type object
        # This breaks certain metrics as accuracy, which makes type_of_target be unknown
        # If while fit a dtype was observed, we try to honor that dtype
        if self.dtype is not None:
            y_inv = y_inv.astype(self.dtype)

        return y_inv

    def is_single_column_target(self) -> bool:
        """Output is encoded with a single column encoding"""
        return self.out_dimensionality == 1

    def _check_data(
        self,
        y: SUPPORTED_TARGET_TYPES,
    ) -> None:
        """
        Perform dimensionality and data type checks on the targets

        Parameters
        ----------
        y: Union[np.ndarray, pd.DataFrame, pd.Series]
            A set of features whose dimensionality and data type is going to be checked
        """
        if not isinstance(
            y, (np.ndarray, pd.DataFrame, List, pd.Series)
        ) and not isinstance(y, spmatrix):
            raise ValueError(
                "Auto-sklearn only supports Numpy arrays, Pandas DataFrames,"
                " pd.Series, sparse data and Python Lists as targets, yet, "
                "the provided input is of type {}".format(type(y))
            )

        if isinstance(y, spmatrix) and not np.issubdtype(y.dtype.type, np.number):
            raise ValueError(
                "When providing a sparse matrix as targets, the only supported "
                "values are numerical. Please consider using a dense"
                " instead."
            )

        if self.data_type is None:
            self.data_type = type(y)
        if self.data_type != type(y):
            self.logger.warning(
                "Auto-sklearn previously received targets of type %s "
                "yet the current features have type %s. Changing the dtype "
                "of inputs to an estimator might cause problems"
                % (
                    str(self.data_type),
                    str(type(y)),
                ),
            )

        # No Nan is supported
        has_nan_values = False
        if hasattr(y, "iloc"):
            has_nan_values = cast(pd.DataFrame, y).isnull().values.any()

        if isinstance(y, spmatrix):
            has_nan_values = not np.array_equal(y.data, y.data)
        else:
            # List and array like values are considered here
            # np.isnan cannot work on strings, so we have to check for every element
            # but NaN, are not equal to themselves:
            has_nan_values = not np.array_equal(y, y)
        if has_nan_values:
            raise ValueError(
                "Target values cannot contain missing/NaN values. "
                "This is not supported by scikit-learn. "
            )

        # Pandas Series is not supported for multi-label indicator
        # This format checks are done by type of target
        try:
            self.type_of_target = type_of_target(y)
        except Exception as e:
            raise ValueError(
                "The provided data could not be interpreted by Sklearn. "
                "While determining the type of the targets via type_of_target "
                "run into exception: {}.".format(e)
            )

        supported_output_types = (
            "binary",
            "continuous",
            "continuous-multioutput",
            "multiclass",
            "multilabel-indicator",
            # Notice unknown/multiclass-multioutput are not supported
            # This can only happen during testing only as estimators
            # should filter out unsupported types.
        )
        if self.type_of_target not in supported_output_types:
            raise ValueError(
                "Provided targets are not supported by Auto-Sklearn. "
                "Provided type is {} whereas supported types are {}.".format(
                    self.type_of_target, supported_output_types
                )
            )

    @property
    def classes_(self) -> np.ndarray:
        """
        Complies with scikit learn classes_ attribute,
        which consist of a ndarray of shape (n_classes,)
        where n_classes are the number of classes seen while fitting
        a encoder to the targets.

        Returns
        -------
        classes_: np.ndarray
            The unique classes seen during encoding of a classifier
        """
        if self.encoder is None:
            return np.array([])
        else:
            return self.encoder.categories_[0]



from typing import Optional

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool, check_none


class ExtraTreesPreprocessorRegression(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        n_estimators,
        criterion,
        min_samples_leaf,
        min_samples_split,
        max_features,
        bootstrap=False,
        max_leaf_nodes=None,
        max_depth="None",
        min_weight_fraction_leaf=0.0,
        oob_score=False,
        n_jobs=1,
        random_state=None,
        verbose=0,
    ):

        self.n_estimators = n_estimators
        self.estimator_increment = 10
        if criterion not in ("mse", "friedman_mse", "mae"):
            raise ValueError(
                "'criterion' is not in ('mse', 'friedman_mse', "
                "'mae'): %s" % criterion
            )
        self.criterion = criterion
        self.min_samples_leaf = min_samples_leaf
        self.min_samples_split = min_samples_split
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.max_leaf_nodes = max_leaf_nodes
        self.max_depth = max_depth
        self.min_weight_fraction_leaf = min_weight_fraction_leaf

        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.preprocessor = None

    def fit(self, X, Y):
        from sklearn.ensemble import ExtraTreesRegressor
        from sklearn.feature_selection import SelectFromModel

        self.n_estimators = int(self.n_estimators)
        self.min_samples_leaf = int(self.min_samples_leaf)
        self.min_samples_split = int(self.min_samples_split)
        self.max_features = float(self.max_features)
        self.bootstrap = check_for_bool(self.bootstrap)
        self.n_jobs = int(self.n_jobs)
        self.verbose = int(self.verbose)

        if check_none(self.max_leaf_nodes):
            self.max_leaf_nodes = None
        else:
            self.max_leaf_nodes = int(self.max_leaf_nodes)

        if check_none(self.max_depth):
            self.max_depth = None
        else:
            self.max_depth = int(self.max_depth)

        self.min_weight_fraction_leaf = float(self.min_weight_fraction_leaf)

        num_features = X.shape[1]
        max_features = int(float(self.max_features) * (np.log(num_features) + 1))
        # Use at most half of the features
        max_features = max(1, min(int(X.shape[1] / 2), max_features))
        estimator = ExtraTreesRegressor(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            bootstrap=self.bootstrap,
            max_features=max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            oob_score=self.oob_score,
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            random_state=self.random_state,
        )

        estimator.fit(X, Y)
        self.preprocessor = SelectFromModel(
            estimator=estimator, threshold="mean", prefit=True
        )

        return self

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "ETR",
            "name": "Extra Trees Regressor Preprocessing",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        n_estimators = Constant("n_estimators", 100)
        criterion = CategoricalHyperparameter(
            "criterion", ["mse", "friedman_mse", "mae"]
        )
        max_features = UniformFloatHyperparameter(
            "max_features", 0.1, 1.0, default_value=1.0
        )

        max_depth = UnParametrizedHyperparameter(name="max_depth", value="None")
        max_leaf_nodes = UnParametrizedHyperparameter("max_leaf_nodes", "None")

        min_samples_split = UniformIntegerHyperparameter(
            "min_samples_split", 2, 20, default_value=2
        )
        min_samples_leaf = UniformIntegerHyperparameter(
            "min_samples_leaf", 1, 20, default_value=1
        )
        min_weight_fraction_leaf = Constant("min_weight_fraction_leaf", 0.0)

        bootstrap = CategoricalHyperparameter(
            "bootstrap", ["True", "False"], default_value="False"
        )

        cs.add_hyperparameters(
            [
                n_estimators,
                criterion,
                max_features,
                max_depth,
                max_leaf_nodes,
                min_samples_split,
                min_samples_leaf,
                min_weight_fraction_leaf,
                bootstrap,
            ]
        )

        return cs



import sklearn.linear_model

from autosklearn.pipeline.components.classification.passive_aggressive import (
    PassiveAggressive,
)

from .test_base import BaseClassificationComponentTest


class PassiveAggressiveComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.98
    res["iris_n_calls"] = 6
    res["default_iris_iterative"] = res["default_iris"]
    res["iris_iterative_n_iter"] = 64
    res["default_iris_proba"] = 0.27840521921952033
    res["default_iris_sparse"] = 0.48
    res["default_digits"] = 0.9162112932604736
    res["digits_n_calls"] = 6
    res["default_digits_iterative"] = res["default_digits"]
    res["digits_iterative_n_iter"] = 64
    res["default_digits_binary"] = 0.99210686095932
    res["default_digits_multilabel"] = 0.910908768565592
    res["default_digits_multilabel_proba"] = 1.0

    sk_mod = sklearn.linear_model.PassiveAggressiveClassifier
    module = PassiveAggressive

    step_hyperparameter = {
        "name": "max_iter",
        "value": module.get_max_iter(),
    }



from __future__ import annotations

from typing import Callable, Union

from smac.callbacks import IncorporateRunResultCallback
from smac.optimizer.smbo import SMBO
from smac.runhistory.runhistory import RunInfo, RunValue

SMACCallback = Callable[[SMBO, RunInfo, RunValue, float], Union[bool, None]]


class SmacRunCallback(IncorporateRunResultCallback):
    def __init__(self, f: SMACCallback):
        self.f = f

    def __call__(
        self,
        smbo: SMBO,
        run_info: RunInfo,
        result: RunValue,
        time_left: float,
    ) -> bool | None:
        """
        Parameters
        ----------
        smbo: SMBO
            The SMAC SMBO object

        run_info: RunInfo
            Information about the run completed

        result: RunValue
            The results of the run

        time_left: float
            How much time is left for the remaining runs

        Returns
        -------
        bool | None
            If False is returned, the optimization loop will stop
        """
        return self.f(smbo, run_info, result, time_left)



from typing import Dict, List, Optional, Tuple, Union, cast

import logging
import warnings

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype, is_sparse
from scipy.sparse import csr_matrix, spmatrix
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError

from autosklearn.util.logging_ import PickableLoggerAdapter

SUPPORTED_FEAT_TYPES = Union[List, pd.DataFrame, np.ndarray, spmatrix]


class FeatureValidator(BaseEstimator):
    """
    Checks the input data to Auto-Sklearn.

    It also determines what columns are categorical and which ones are numerical,
    so that the pre-processing pipeline can process this columns accordingly.

    Attributes
    ----------
    feat_type: Optional[List[str]] = None
        In case the dataset is not a pandas DataFrame:
            +   If provided, this list indicates which columns should be treated as
                categorical it is internally transformed into a dictionary that
                indicates a mapping from column index to categorical/numerical.
            +   If not provided, by default all columns are treated as numerical

        If the input dataset is of type pandas dataframe, this argument
        must be none, as the column type will be inferred from the pandas dtypes.

    data_type:
        Class name of the data type provided during fit.
    """

    def __init__(
        self,
        feat_type: Optional[List[str]] = None,
        logger: Optional[PickableLoggerAdapter] = None,
        allow_string_features: bool = True,
    ) -> None:
        # If a dataframe was provided, we populate
        # this attribute with a mapping from column to
        # {numerical | categorical | string}
        self.feat_type: Optional[Dict[Union[str, int], str]] = None
        if feat_type is not None:
            if isinstance(feat_type, dict):
                self.feat_type = feat_type
            elif not isinstance(feat_type, List):
                raise ValueError(
                    "Auto-Sklearn expects a list of categorical/"
                    "numerical/string feature types, yet a"
                    " {} was provided".format(type(feat_type))
                )
            else:

                # Convert to a dictionary which will be passed to the ColumnTransformer
                # Column Transformer supports strings or integer indexes
                self.feat_type = {i: feat for i, feat in enumerate(feat_type)}

        # Register types to detect unsupported data format changes
        self.data_type = None  # type: Optional[type]
        self.dtypes = {}  # type: Dict[str, str]

        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self._is_fitted = False
        self.allow_string_features = allow_string_features

    def fit(
        self,
        X_train: SUPPORTED_FEAT_TYPES,
        X_test: Optional[SUPPORTED_FEAT_TYPES] = None,
    ) -> "FeatureValidator":
        """
        Validates input data to Auto-Sklearn.
        The supported data types are List, numpy arrays and pandas DataFrames.
        CSR sparse data types are also supported

        Parameters
        ----------
        X_train: SUPPORTED_FEAT_TYPES
            A set of features that are going to be validated (type and dimensionality
            checks) and a encoder fitted in the case the data needs encoding
        X_test: Optional[SUPPORTED_FEAT_TYPES]
            A hold out set of data used for checking
        """
        # If a list was provided, it will be converted to pandas
        if isinstance(X_train, List):
            X_train, X_test = self.list_to_dataframe(X_train, X_test)

        self._check_data(X_train)

        # Handle categorical feature identification for the pipeline
        if hasattr(X_train, "iloc"):
            if self.feat_type is not None:
                raise ValueError(
                    "When providing a DataFrame to Auto-Sklearn, we extract "
                    "the feature types from the DataFrame.dtypes. That is, "
                    "providing the option feat_type to the fit method is not "
                    "supported when using a Dataframe. Please make sure that the "
                    "type of each column in your DataFrame is properly set. "
                    "More details about having the correct data type in your "
                    "DataFrame can be seen in "
                    "https://pandas.pydata.org/pandas-docs/stable/reference"
                    "/api/pandas.DataFrame.astype.html"
                )
            else:
                self.feat_type = self.get_feat_type_from_columns(X_train)
        else:
            # Numpy array was provided
            if self.feat_type is None:
                # Assume numerical columns if a numpy array has no feature types
                self.feat_type = {i: "numerical" for i in range(np.shape(X_train)[1])}
            else:
                # Check The feat type provided
                if len(self.feat_type) != np.shape(X_train)[1]:
                    raise ValueError(
                        "Array feat_type does not have same number of "
                        "variables as X has features. %d vs %d."
                        % (len(self.feat_type), np.shape(X_train)[1])
                    )
                if not all([isinstance(f, str) for f in self.feat_type.values()]):
                    raise ValueError(
                        "feat_type must only contain strings: {}".format(
                            list(self.feat_type.values()),
                        )
                    )

                for ft in self.feat_type.values():
                    if ft.lower() not in ["categorical", "numerical", "string"]:
                        raise ValueError(
                            "Only `Categorical`, `Numerical` and `String` are "
                            "valid feature types"
                        )

        if X_test is not None:
            self._check_data(X_test)

            if np.shape(X_train)[1] != np.shape(X_test)[1]:
                raise ValueError(
                    "The feature dimensionality of the train and test "
                    "data does not match train({}) != test({})".format(
                        np.shape(X_train)[1], np.shape(X_test)[1]
                    )
                )

        self._is_fitted = True

        return self

    def transform(
        self,
        X: SUPPORTED_FEAT_TYPES,
    ) -> Union[np.ndarray, spmatrix, pd.DataFrame]:
        """
        Validates and fit a categorical encoder (if needed) to the features.
        The supported data types are List, numpy arrays and pandas DataFrames.

        Parameters
        ----------
        X_train: SUPPORTED_FEAT_TYPES
            A set of features, whose categorical features are going to be
            transformed

        Return
        ------
        np.ndarray | spmatrix | pd.DataFrame:
            The transformed array
        """
        if not self._is_fitted:
            raise NotFittedError(
                "Cannot call transform on a validator that is not fitted"
            )

        # If a list was provided, it will be converted to pandas
        if isinstance(X, List):
            X_transformed, _ = self.list_to_dataframe(X)
        else:
            X_transformed = X

        # Check the data here so we catch problems on new test data
        self._check_data(X_transformed)

        # Sparse related transformations
        # Not all sparse format support index sorting
        if isinstance(X_transformed, spmatrix):
            if not isinstance(X_transformed, csr_matrix):
                self.logger.warning(
                    f"Sparse data provided is of type {type(X_transformed)} "
                    "yet Auto-Sklearn only support csr_matrix. Auto-sklearn "
                    "will convert the provided data to the csr_matrix format."
                )
                X_transformed = X_transformed.tocsr(copy=False)

            X_transformed.sort_indices()

        return X_transformed

    def _check_data(
        self,
        X: SUPPORTED_FEAT_TYPES,
    ) -> None:
        """
        Feature dimensionality and data type checks

        Parameters
        ----------
        X: SUPPORTED_FEAT_TYPES
            A set of features that are going to be validated (type and dimensionality)
            and a encoder fitted in the case the data needs encoding
        """
        # We consider columns that are all nan in a pandas frame as category
        if hasattr(X, "columns"):
            for column in cast(pd.DataFrame, X).columns:
                if X[column].isna().all():
                    X[column] = X[column].astype("category")

        if not isinstance(X, (np.ndarray, pd.DataFrame)) and not isinstance(
            X, spmatrix
        ):
            raise ValueError(
                "Auto-sklearn only supports Numpy arrays, Pandas DataFrames,"
                " scipy sparse and Python Lists, yet, the provided input is"
                " of type {}".format(type(X))
            )

        if self.data_type is None:
            self.data_type = type(X)

        if self.data_type != type(X):
            self.logger.warning(
                f"Auto-sklearn previously received features of type {self.data_type} "
                f"yet the current features have type {type(X)}. Changing the dtype "
                "of inputs to an estimator might cause problems"
            )

        # Do not support category/string numpy data. Only numbers
        if hasattr(X, "dtype"):
            if not np.issubdtype(X.dtype.type, np.number):  # type: ignore[union-attr]
                raise ValueError(
                    "When providing a numpy array to Auto-sklearn, the only valid"
                    f" dtypes are numerical ones. The provided data type {X.dtype.type}"
                    " is not supported."
                )

        # Then for Pandas, we do not support Nan in categorical columns
        if hasattr(X, "iloc"):
            # If entered here, we have a pandas dataframe
            X = cast(pd.DataFrame, X)

            dtypes = {col: X[col].dtype.name.lower() for col in X.columns}
            if len(self.dtypes) > 0:
                if self.dtypes != dtypes:
                    # To support list, we need to support object inference.
                    # In extreme cases, the train column might be all integer,
                    # and the test column might be float.
                    self.logger.warning(
                        "Changing the dtype of the features after fit() is "
                        "not recommended. Fit() method was called with "
                        "{} whereas the new features have {} as type".format(
                            self.dtypes,
                            dtypes,
                        )
                    )
            else:
                self.dtypes = dtypes

    def get_feat_type_from_columns(
        self,
        X: pd.DataFrame,
    ) -> Dict[Union[str, int], str]:
        """
        Returns a dictionary that maps pandas dataframe columns to a feature type.
        This feature type can be categorical, numerical or string

        Parameters
        ----------
        X: pd.DataFrame
            A set of features that are going to be validated (type and dimensionality
            checks) and a encoder fitted in the case the data needs encoding

        Returns
        -------
        feat_type:
            dictionary with column to feature type mapping
        """
        # Also, register the feature types for the estimator
        feat_type = {}

        # Make sure each column is a valid type
        for i, column in enumerate(X.columns):
            if is_sparse(X[column]):
                raise ValueError(
                    "Auto-sklearn does not yet support sparse pandas Series."
                    f" Please convert {column} to a dense format."
                )
            elif X[column].dtype.name in ["category", "bool"]:
                feat_type[column] = "categorical"
            elif X[column].dtype.name == "string":
                if self.allow_string_features:
                    feat_type[column] = "string"
                else:
                    feat_type[column] = "categorical"
                    warnings.warn(
                        f"you disabled text encoding column {column} will be "
                        f"encoded as category"
                    )
            # Move away from np.issubdtype as it causes
            # TypeError: data type not understood in certain pandas types
            elif not is_numeric_dtype(X[column]):
                if X[column].dtype.name == "object":
                    warnings.warn(
                        f"Input Column {column} has generic type object. "
                        f"Autosklearn will treat this column as string. "
                        f"Please ensure that this setting is suitable for your task.",
                        UserWarning,
                    )
                    if self.allow_string_features:
                        feat_type[column] = "string"
                    else:
                        feat_type[column] = "categorical"
                        warnings.warn(
                            f"you disabled text encoding column {column} will be"
                            f"encoded as category"
                        )
                elif pd.core.dtypes.common.is_datetime_or_timedelta_dtype(
                    X[column].dtype
                ):
                    raise ValueError(
                        "Auto-sklearn does not support time and/or date datatype as "
                        f"given in column {column}. Please convert the time "
                        " information to a numerical value first. One example on how to"
                        " do this can be found on "
                        " https://stats.stackexchange.com/questions/311494/"
                    )
                else:
                    raise ValueError(
                        "Input Column {} has unsupported dtype {}. "
                        "Supported column types are categorical/bool/numerical dtypes. "
                        "Make sure your data is formatted in a correct way, "
                        "before feeding it to Auto-Sklearn.".format(
                            column,
                            X[column].dtype.name,
                        )
                    )
            else:
                feat_type[column] = "numerical"
        return feat_type

    def list_to_dataframe(
        self,
        X_train: SUPPORTED_FEAT_TYPES,
        X_test: Optional[SUPPORTED_FEAT_TYPES] = None,
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        """
        Converts a list to a DataFrame. In this process, column types are inferred.

        If test data is provided, we proactively match it to train data

        Parameters
        ----------
        X_train: SUPPORTED_FEAT_TYPES
            A set of features that are going to be validated (type and dimensionality
            checks) and a encoder fitted in the case the data needs encoding

        X_test: Optional[SUPPORTED_FEAT_TYPES]
            A hold out set of data used for checking

        Returns
        -------
        Union[pd.DataFrame, pd.DataFrame]:
            transformed (train, test) data from list to pandas DataFrame
        """
        # If a list was provided, it will be converted to pandas
        X_train = pd.DataFrame(data=X_train).convert_dtypes()

        # Store the dtypes and use in case of re-fit
        if len(self.dtypes) == 0:
            self.dtypes = {
                col: X_train[col].dtype.name.lower() for col in X_train.columns
            }
        else:
            for col in X_train.columns:
                # Try to convert to the original dtype used to fit the validator
                # But also be robust to extreme cases (for example, the train data for a
                # column was all np.int-like and the test data is np.float-type)
                try:
                    X_train[col] = X_train[col].astype(self.dtypes[col])
                except Exception as e:
                    self.logger.warning(
                        f"Failed to format column {col} as {self.dtypes[col]}: {e}"
                    )
                    self.dtypes[col] = X_train[col].dtype.name.lower()

        self.logger.warning(
            "The provided feature types to autosklearn are of type list."
            "Features have been interpreted as: {}".format(
                [(col, t) for col, t in zip(X_train.columns, X_train.dtypes)]
            )
        )
        if X_test is not None:
            if not isinstance(X_test, List):
                self.logger.warning(
                    "Train features are a list while the provided test data"
                    "is {}. X_test will be casted as DataFrame.".format(type(X_test))
                )
            X_test = pd.DataFrame(data=X_test)
            for col in X_test.columns:
                try:
                    X_test[col] = X_test[col].astype(self.dtypes[col])
                except Exception as e:
                    self.logger.warning(
                        f"Failed to format column {col} as {self.dtypes[col]}: {e}"
                    )
                    self.dtypes[col] = X_test[col].dtype.name.lower()

        return X_train, X_test



import logging
import logging.handlers
import os
import shutil
import sys
import time

import numpy as np
import pynisher
from smac.runhistory.runhistory import RunInfo
from smac.stats.stats import Stats
from smac.tae import StatusType
from smac.utils.constants import MAXINT

from autosklearn.evaluation import ExecuteTaFuncWithQueue, get_cost_of_crash
from autosklearn.metrics import accuracy, log_loss

import unittest
import unittest.mock

this_directory = os.path.dirname(__file__)
sys.path.append(this_directory)
from evaluation_util import (  # noqa E402
    get_evaluation_backend,
    get_multiclass_classification_datamanager,
)


def safe_eval_success_mock(*args, **kwargs):
    queue = kwargs["queue"]
    queue.put({"status": StatusType.SUCCESS, "loss": 0.5, "additional_run_info": ""})


class EvaluationTest(unittest.TestCase):
    def setUp(self):
        self.datamanager = get_multiclass_classification_datamanager()
        self.tmp = os.path.join(os.getcwd(), ".test_evaluation")
        self.logger = logging.getLogger()
        scenario_mock = unittest.mock.Mock()
        scenario_mock.wallclock_limit = 10
        scenario_mock.algo_runs_timelimit = 1000
        scenario_mock.ta_run_limit = 100
        self.scenario = scenario_mock
        stats = Stats(scenario_mock)
        stats.start_timing()
        self.stats = stats
        self.logger_port = logging.handlers.DEFAULT_TCP_LOGGING_PORT

        self.backend = get_evaluation_backend()

        try:
            shutil.rmtree(self.tmp)
        except Exception:
            pass

    def tearDown(self):
        try:
            shutil.rmtree(self.tmp)
        except Exception:
            pass

    ############################################################################
    # pynisher tests
    def test_pynisher_memory_error(self):
        def fill_memory():
            a = np.random.random_sample((10000, 10000)).astype(np.float64)
            return np.sum(a)

        safe_eval = pynisher.enforce_limits(mem_in_mb=1)(fill_memory)
        safe_eval()
        self.assertEqual(safe_eval.exit_status, pynisher.MemorylimitException)

    def test_pynisher_timeout(self):
        def run_over_time():
            time.sleep(2)

        safe_eval = pynisher.enforce_limits(wall_time_in_s=1, grace_period_in_s=0)(
            run_over_time
        )
        safe_eval()
        self.assertEqual(safe_eval.exit_status, pynisher.TimeoutException)

    ############################################################################
    # Test ExecuteTaFuncWithQueue.run_wrapper()
    @unittest.mock.patch("autosklearn.evaluation.train_evaluator.eval_holdout")
    def test_eval_with_limits_holdout(self, pynisher_mock):
        pynisher_mock.side_effect = safe_eval_success_mock
        config = unittest.mock.Mock()
        config.config_id = 198
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            resampling_strategy="holdout",
            multi_objectives=["cost"],
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="fork",
        )
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[0].config.config_id, 198)
        self.assertEqual(info[1].status, StatusType.SUCCESS)
        self.assertEqual(info[1].cost, 0.5)
        self.assertIsInstance(info[1].time, float)

    @unittest.mock.patch("pynisher.enforce_limits")
    def test_zero_or_negative_cutoff(self, pynisher_mock):
        config = unittest.mock.Mock()
        config.config_id = 198
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="forkserver",
        )
        self.scenario.wallclock_limit = 5
        self.stats.submitted_ta_runs += 1
        run_info, run_value = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=9,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(run_value.status, StatusType.STOP)

    @unittest.mock.patch("pynisher.enforce_limits")
    def test_cutoff_lower_than_remaining_time(self, pynisher_mock):
        config = unittest.mock.Mock()
        config.config_id = 198
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="forkserver",
        )
        self.stats.ta_runs = 1
        ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(pynisher_mock.call_args[1]["wall_time_in_s"], 4)
        self.assertIsInstance(pynisher_mock.call_args[1]["wall_time_in_s"], int)

    @unittest.mock.patch("autosklearn.evaluation.train_evaluator.eval_holdout")
    def test_eval_with_limits_holdout_fail_silent(self, pynisher_mock):
        pynisher_mock.return_value = None
        config = unittest.mock.Mock()
        config.origin = "MOCK"
        config.config_id = 198
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="fork",
        )

        # The following should not fail because abort on first config crashed is false
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=60,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.CRASHED)
        self.assertEqual(info[1].cost, 1.0)
        self.assertIsInstance(info[1].time, float)
        self.assertEqual(
            info[1].additional_info,
            {
                "configuration_origin": "MOCK",
                "error": "Result queue is empty",
                "exit_status": 0,
                "exitcode": 0,
                "subprocess_stdout": "",
                "subprocess_stderr": "",
            },
        )

        self.stats.submitted_ta_runs += 1
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.CRASHED)
        self.assertEqual(info[1].cost, 1.0)
        self.assertIsInstance(info[1].time, float)
        self.assertEqual(
            info[1].additional_info,
            {
                "configuration_origin": "MOCK",
                "error": "Result queue is empty",
                "exit_status": 0,
                "exitcode": 0,
                "subprocess_stdout": "",
                "subprocess_stderr": "",
            },
        )

    @unittest.mock.patch("autosklearn.evaluation.train_evaluator.eval_holdout")
    def test_eval_with_limits_holdout_fail_memory_error(self, pynisher_mock):
        pynisher_mock.side_effect = MemoryError
        config = unittest.mock.Mock()
        config.config_id = 198
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[log_loss],
            cost_for_crash=get_cost_of_crash([log_loss]),
            abort_on_first_run_crash=False,
            pynisher_context="fork",
        )
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.MEMOUT)

        # For logloss, worst possible result is MAXINT
        worst_possible_result = MAXINT
        self.assertEqual(info[1].cost, worst_possible_result)
        self.assertIsInstance(info[1].time, float)
        self.assertNotIn("exitcode", info[1].additional_info)

    @unittest.mock.patch("pynisher.enforce_limits")
    def test_eval_with_limits_holdout_fail_timeout(self, pynisher_mock):
        config = unittest.mock.Mock()
        config.config_id = 198

        m1 = unittest.mock.Mock()
        m2 = unittest.mock.Mock()
        m1.return_value = m2
        pynisher_mock.return_value = m1
        m2.exit_status = pynisher.TimeoutException
        m2.wall_clock_time = 30
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="forkserver",
        )
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.TIMEOUT)
        self.assertEqual(info[1].cost, 1.0)
        self.assertIsInstance(info[1].time, float)
        self.assertNotIn("exitcode", info[1].additional_info)

    @unittest.mock.patch("pynisher.enforce_limits")
    def test_eval_with_limits_holdout_timeout_with_results_in_queue(
        self, pynisher_mock
    ):
        config = unittest.mock.Mock()
        config.config_id = 198

        def side_effect(**kwargs):
            queue = kwargs["queue"]
            queue.put(
                {"status": StatusType.SUCCESS, "loss": 0.5, "additional_run_info": {}}
            )

        m1 = unittest.mock.Mock()
        m2 = unittest.mock.Mock()
        m1.return_value = m2
        pynisher_mock.return_value = m1
        m2.side_effect = side_effect
        m2.exit_status = pynisher.TimeoutException
        m2.wall_clock_time = 30

        # Test for a succesful run
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="forkserver",
        )
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.SUCCESS)
        self.assertEqual(info[1].cost, 0.5)
        self.assertIsInstance(info[1].time, float)
        self.assertNotIn("exitcode", info[1].additional_info)

        # And a crashed run which is in the queue
        def side_effect(**kwargs):
            queue = kwargs["queue"]
            queue.put(
                {"status": StatusType.CRASHED, "loss": 2.0, "additional_run_info": {}}
            )

        m2.side_effect = side_effect
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="forkserver",
        )
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.CRASHED)
        self.assertEqual(info[1].cost, 1.0)
        self.assertIsInstance(info[1].time, float)
        self.assertNotIn("exitcode", info[1].additional_info)

    @unittest.mock.patch("autosklearn.evaluation.train_evaluator.eval_holdout")
    def test_eval_with_limits_holdout_2(self, eval_houldout_mock):
        config = unittest.mock.Mock()
        config.config_id = 198

        def side_effect(*args, **kwargs):
            queue = kwargs["queue"]
            queue.put(
                {
                    "status": StatusType.SUCCESS,
                    "loss": 0.5,
                    "additional_run_info": kwargs["instance"],
                }
            )

        eval_houldout_mock.side_effect = side_effect
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="fork",
        )
        self.scenario.wallclock_limit = 180
        instance = "{'subsample': 30}"
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=instance,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.SUCCESS)
        self.assertEqual(len(info[1].additional_info), 2)
        self.assertIn("configuration_origin", info[1].additional_info)
        self.assertEqual(info[1].additional_info["message"], "{'subsample': 30}")

    @unittest.mock.patch("autosklearn.evaluation.train_evaluator.eval_holdout")
    def test_exception_in_target_function(self, eval_holdout_mock):
        config = unittest.mock.Mock()
        config.config_id = 198

        eval_holdout_mock.side_effect = ValueError
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            autosklearn_seed=1,
            port=self.logger_port,
            multi_objectives=["cost"],
            resampling_strategy="holdout",
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            pynisher_context="fork",
        )
        self.stats.submitted_ta_runs += 1
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=30,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(info[1].status, StatusType.CRASHED)
        self.assertEqual(info[1].cost, 1.0)
        self.assertIsInstance(info[1].time, float)
        self.assertEqual(info[1].additional_info["error"], "ValueError()")
        self.assertIn("traceback", info[1].additional_info)
        self.assertNotIn("exitcode", info[1].additional_info)

    def test_silent_exception_in_target_function(self):
        config = unittest.mock.Mock()
        config.config_id = 198

        delattr(self.backend, "save_additional_data")
        ta = ExecuteTaFuncWithQueue(
            backend=self.backend,
            port=self.logger_port,
            autosklearn_seed=1,
            resampling_strategy="holdout",
            multi_objectives=["cost"],
            stats=self.stats,
            memory_limit=3072,
            metrics=[accuracy],
            cost_for_crash=get_cost_of_crash([accuracy]),
            abort_on_first_run_crash=False,
            iterative=False,
            pynisher_context="fork",
        )
        ta.pynisher_logger = unittest.mock.Mock()
        self.stats.submitted_ta_runs += 1
        info = ta.run_wrapper(
            RunInfo(
                config=config,
                cutoff=3000,
                instance=None,
                instance_specific=None,
                seed=1,
                capped=False,
            )
        )
        self.assertEqual(
            info[1].status, StatusType.CRASHED, msg=str(info[1].additional_info)
        )
        self.assertEqual(info[1].cost, 1.0)
        self.assertIsInstance(info[1].time, float)
        self.assertIn(
            info[1].additional_info["error"],
            (
                """AttributeError("'BackendMock' object has no attribute """
                """'save_additional_data'",)""",
                """AttributeError("'BackendMock' object has no attribute """
                """'save_additional_data'")""",
                """AttributeError('save_additional_data')""",
            ),
        )
        self.assertNotIn("exitcode", info[1].additional_info)
        self.assertNotIn("exit_status", info[1].additional_info)
        self.assertNotIn("traceback", info[1])



import sklearn.tree

from autosklearn.pipeline.components.regression.decision_tree import DecisionTree

from .test_base import BaseRegressionComponentTest


class DecisionTreeComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.35616796434879905
    res["default_boston_iterative"] = None
    res["default_boston_sparse"] = 0.18031669797027394
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = 0.1564592449511697
    res["default_diabetes_iterative"] = None
    res["default_diabetes_sparse"] = -0.020818312539637507
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.tree.DecisionTreeRegressor

    module = DecisionTree



"""
This file tests the ensemble builder with real runs generated from running AutoML
"""
from __future__ import annotations

from typing import Callable

from autosklearn.automl import AutoML
from autosklearn.ensemble_building.builder import EnsembleBuilder

from pytest_cases import parametrize_with_cases
from pytest_cases.filters import has_tag
from unittest.mock import MagicMock, patch

import test.test_automl.cases as cases


@parametrize_with_cases(
    "automl",
    cases=cases,
    filter=has_tag("fitted") & ~has_tag("no_ensemble"),
)
def case_real_runs(
    automl: AutoML,
    make_ensemble_builder: Callable[..., EnsembleBuilder],
) -> EnsembleBuilder:
    """Uses real runs from a fitted automl instance which have an ensemble

    This will copy the ensemble builder based on the AutoML instance parameterss. This
    includes ensemble_nbest, ensemble_size, etc...
    """
    builder = make_ensemble_builder(automl=automl)
    return builder


@parametrize_with_cases("builder", cases=case_real_runs)
def test_run_builds_valid_ensemble(builder: EnsembleBuilder) -> None:
    """
    Expects
    -------
    * Using the same builder as used in the real run should result in the same
      candidate models for the ensemble.
    * Check that there is no overlap between candidate models and those deleted
    * The generated ensemble should not be empty
    * If any deleted, should be no overlap with those deleted and those in ensemble
    * If any deleted, they should all be worse than those in the ensemble
    """
    # We need to clear previous candidates so the ensemble builder is presented with
    # only "new" runs and has no information of previous candidates
    if builder.previous_candidates_path.exists():
        builder.previous_candidates_path.unlink()

    # So we can capture the saved ensemble
    mock_save = MagicMock()
    builder.backend.save_ensemble = mock_save  # type: ignore

    # So we can capture what was deleted
    mock_delete = MagicMock()
    builder.delete_runs = mock_delete  # type: ignore

    # So we can capture the candidate runs used, we still wrap the actual fitting
    with patch(
        "autosklearn.ensembles.ensemble_selection.EnsembleSelection"
        ".get_validation_performance"
    ) as mock_get_validation_performance:
        mock_get_validation_performance.return_value = 0.2
        with patch.object(
            builder, "fit_ensemble", wraps=builder.fit_ensemble
        ) as mock_fit:
            history, nbest = builder.main()

    # Check the ensemble was fitted once
    mock_save.assert_called_once()
    _, kwargs = mock_save.call_args
    ens = kwargs["ensemble"]  # `backend.save_ensemble(ens, ...)`
    ensemble_ids = set(ens.get_selected_model_identifiers())
    assert len(ensemble_ids) > 0

    assert mock_fit.call_count == 1
    # Check that the ids of runs in the ensemble were all candidates
    candidates = mock_fit.call_args[1]["candidates"]
    candidate_ids = {run.id for run in candidates}
    assert ensemble_ids <= candidate_ids

    # Could be the case no run is deleted
    if not mock_delete.called:
        return

    args, _ = mock_delete.call_args
    deleted = args[0]  # `delete_runs(runs)`

    # If we deleted runs, we better make sure of a few things
    if len(deleted) > 0:
        deleted_ids = {run.id for run in deleted}

        # Make sure theres no overlap between candidates/ensemble and those deleted
        assert not any(deleted_ids & candidate_ids)
        assert not any(deleted_ids & ensemble_ids)

        # Make sure that the best deleted model is still worse than the worst candidate
        # This does not necessarily hold with respect to any single metric in the
        # multiobjective case as the best deleted may have been deleted with respect to
        # some other metric when compared to the worst candidate, we can't know which.
        if len(builder.metrics) == 1:
            metric = builder.metrics[0]
            best_deleted = min(
                deleted, key=lambda r: (r.losses[metric.name], r.num_run)
            )
            worst_candidate = max(
                candidates, key=lambda r: (r.losses[metric.name], r.num_run)
            )

            a = (worst_candidate.losses[metric.name], worst_candidate.num_run)
            b = (best_deleted.losses[metric.name], best_deleted.num_run)
            assert a <= b
